#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""MODIS L1b and L2 test fixtures."""
from __future__ import annotations

from datetime import datetime, timedelta
from typing import Optional

import numpy as np
import pytest
from pyhdf.SD import SD, SDC

# Level 1 Fixtures

AVAILABLE_1KM_VIS_PRODUCT_NAMES = [str(x) for x in range(8, 13)]
AVAILABLE_1KM_VIS_PRODUCT_NAMES += ['13lo', '13hi', '14lo', '14hi']
AVAILABLE_1KM_VIS_PRODUCT_NAMES += [str(x) for x in range(15, 20)]
AVAILABLE_1KM_IR_PRODUCT_NAMES = [str(x) for x in range(20, 37)]
AVAILABLE_1KM_PRODUCT_NAMES = AVAILABLE_1KM_VIS_PRODUCT_NAMES + AVAILABLE_1KM_IR_PRODUCT_NAMES
AVAILABLE_HKM_PRODUCT_NAMES = [str(x) for x in range(3, 8)]
AVAILABLE_QKM_PRODUCT_NAMES = ['1', '2']
SCAN_LEN_5KM = 406
SCAN_WIDTH_5KM = 270
SCALE_FACTOR = 1
RES_TO_REPEAT_FACTOR = {
    250: 20,
    500: 10,
    1000: 5,
    5000: 1,
}


def _shape_for_resolution(resolution: int) -> tuple[int, int]:
    assert resolution in RES_TO_REPEAT_FACTOR
    factor = RES_TO_REPEAT_FACTOR[resolution]
    if factor == 1:
        return SCAN_LEN_5KM, SCAN_WIDTH_5KM

    factor_1km = RES_TO_REPEAT_FACTOR[1000]
    shape_1km = (factor_1km * SCAN_LEN_5KM, factor_1km * SCAN_WIDTH_5KM + 4)
    factor //= 5
    return factor * shape_1km[0], factor * shape_1km[1]


def _generate_lonlat_data(resolution: int) -> np.ndarray:
    shape = _shape_for_resolution(resolution)
    lat = np.repeat(np.linspace(35., 45., shape[0])[:, None], shape[1], 1)
    lat *= np.linspace(0.9, 1.1, shape[1])
    lon = np.repeat(np.linspace(-45., -35., shape[1])[None, :], shape[0], 0)
    lon *= np.linspace(0.9, 1.1, shape[0])[:, None]
    return lon.astype(np.float32), lat.astype(np.float32)


def _generate_angle_data(resolution: int) -> np.ndarray:
    shape = _shape_for_resolution(resolution)
    data = np.repeat(abs(np.linspace(-65.2, 65.4, shape[1]))[None, :], shape[0], 0)
    return (data * 100).astype(np.int16)


def _generate_visible_data(resolution: int, num_bands: int, dtype=np.uint16) -> np.ndarray:
    shape = _shape_for_resolution(resolution)
    data = np.zeros((num_bands, shape[0], shape[1]), dtype=dtype)
    return data


def _get_lonlat_variable_info(resolution: int) -> dict:
    lon_5km, lat_5km = _generate_lonlat_data(resolution)
    return {
        'Latitude': {'data': lat_5km,
                     'type': SDC.FLOAT32,
                     'fill_value': -999,
                     'attrs': {'dim_labels': ['Cell_Along_Swath_5km:mod35', 'Cell_Across_Swath_5km:mod35']}},
        'Longitude': {'data': lon_5km,
                      'type': SDC.FLOAT32,
                      'fill_value': -999,
                      'attrs': {'dim_labels': ['Cell_Along_Swath_5km:mod35', 'Cell_Across_Swath_5km:mod35']}},
    }


def _get_angles_variable_info(resolution: int) -> dict:
    angle_data = _generate_angle_data(resolution)
    dim_factor = RES_TO_REPEAT_FACTOR[resolution] * 2
    angle_info = {
        'data': angle_data,
        'type': SDC.INT16,
        'fill_value': -32767,
        'attrs': {
            'dim_labels': [
                f'{dim_factor}*nscans:MODIS_SWATH_Type_L1B',
                '1KM_geo_dim:MODIS_SWATH_Type_L1B'],
            'scale_factor': 0.01
        },
    }
    angles_info = {}
    for var_name in ('SensorAzimuth', 'SensorZenith', 'SolarAzimuth', 'SolarZenith'):
        angles_info[var_name] = angle_info
    return angles_info


def _get_visible_variable_info(var_name: str, resolution: int, bands: list[str]):
    num_bands = len(bands)
    data = _generate_visible_data(resolution, len(bands))
    dim_factor = RES_TO_REPEAT_FACTOR[resolution] * 2
    band_dim_name = f"Band_{resolution}_{num_bands}_RefSB:MODIS_SWATH_Type_L1B"
    row_dim_name = f'{dim_factor}*nscans:MODIS_SWATH_Type_L1B'
    col_dim_name = 'Max_EV_frames:MODIS_SWATH_Type_L1B'
    return {
        var_name: {
            'data': data,
            'type': SDC.UINT16,
            'fill_value': 0,
            'attrs': {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                'dim_labels': [band_dim_name,
                               row_dim_name,
                               col_dim_name],
                'valid_range': (0, 32767),
                'reflectance_scales': (1,) * num_bands,
                'reflectance_offsets': (0,) * num_bands,
                'band_names': ",".join(bands),
            },
        },
        var_name + '_Uncert_Indexes': {
            'data': np.zeros(data.shape, dtype=np.uint8),
            'type': SDC.UINT8,
            'fill_value': 255,
            'attrs': {
                'dim_labels': [band_dim_name,
                               row_dim_name,
                               col_dim_name],
            },
        },
    }


def _get_emissive_variable_info(var_name: str, resolution: int, bands: list[str]):
    num_bands = len(bands)
    data = _generate_visible_data(resolution, len(bands))
    dim_factor = RES_TO_REPEAT_FACTOR[resolution] * 2
    band_dim_name = f"Band_{resolution}_{num_bands}_Emissive:MODIS_SWATH_Type_L1B"
    row_dim_name = f'{dim_factor}*nscans:MODIS_SWATH_Type_L1B'
    col_dim_name = 'Max_EV_frames:MODIS_SWATH_Type_L1B'
    return {
        var_name: {
            'data': data,
            'type': SDC.UINT16,
            'fill_value': 0,
            'attrs': {
                'dim_labels': [band_dim_name,
                               row_dim_name,
                               col_dim_name],
                'valid_range': (0, 32767),
                'band_names': ",".join(bands),
            },
        },
        var_name + '_Uncert_Indexes': {
            'data': np.zeros(data.shape, dtype=np.uint8),
            'type': SDC.UINT8,
            'fill_value': 255,
            'attrs': {
                'dim_labels': [band_dim_name,
                               row_dim_name,
                               col_dim_name],
            },
        },
    }


def _get_l1b_geo_variable_info(filename: str,
                               geo_resolution: int,
                               include_angles: bool = True
                               ) -> dict:
    variables_info = {}
    variables_info.update(_get_lonlat_variable_info(geo_resolution))
    if include_angles:
        variables_info.update(_get_angles_variable_info(geo_resolution))
    return variables_info


def generate_nasa_l1b_filename(prefix):
    """Generate a filename that follows NASA MODIS L1b convention."""
    now = datetime.now()
    return f'{prefix}_A{now:%y%j_%H%M%S}_{now:%Y%j%H%M%S}.hdf'


def generate_imapp_filename(suffix):
    """Generate a filename that follows IMAPP MODIS L1b convention."""
    now = datetime.now()
    return f't1.{now:%y%j.%H%M}.{suffix}.hdf'


def create_hdfeos_test_file(filename: str,
                            variable_infos: dict,
                            geo_resolution: Optional[int] = None,
                            file_shortname: Optional[str] = None,
                            include_metadata: bool = True):
    """Create a fake MODIS L1b HDF4 file with headers.

    Args:
        filename: Full path of filename to be created.
        variable_infos: Dictionary mapping HDF4 variable names to dictionary
            of variable information (see ``_add_variable_to_file``).
        geo_resolution: Resolution of geolocation datasets to be stored in the
            metadata strings stored in the global metadata attributes. Only
            used if ``include_metadata`` is ``True`` (default).
        file_shortname: Short name of the file to be stored in global metadata
            attributes. Only used if ``include_metadata`` is ``True``
            (default).
        include_metadata: Include global metadata attributes (default: True).

    """
    h = SD(filename, SDC.WRITE | SDC.CREATE)

    if include_metadata:
        if geo_resolution is None or file_shortname is None:
            raise ValueError("'geo_resolution' and 'file_shortname' are required when including metadata.")
        setattr(h, 'CoreMetadata.0', _create_core_metadata(file_shortname))  # noqa
        setattr(h, 'StructMetadata.0', _create_struct_metadata(geo_resolution))  # noqa
        setattr(h, 'ArchiveMetadata.0', _create_header_metadata())  # noqa

    for var_name, var_info in variable_infos.items():
        _add_variable_to_file(h, var_name, var_info)

    h.end()


def _add_variable_to_file(h, var_name, var_info):
    v = h.create(var_name, var_info['type'], var_info['data'].shape)
    v[:] = var_info['data']
    dim_count = 0
    for dimension_name in var_info['attrs']['dim_labels']:
        v.dim(dim_count).setname(dimension_name)
        dim_count += 1
    v.setfillvalue(var_info['fill_value'])
    v.scale_factor = var_info['attrs'].get('scale_factor', SCALE_FACTOR)
    for attr_key, attr_val in var_info['attrs'].items():
        if attr_key == 'dim_labels':
            continue
        setattr(v, attr_key, attr_val)


def _create_core_metadata(file_shortname: str) -> str:
    beginning_date = datetime.now()
    ending_date = beginning_date + timedelta(minutes=5)
    core_metadata_header = "GROUP = INVENTORYMETADATA\nGROUPTYPE = MASTERGROUP\n\n" \
                           "GROUP = RANGEDATETIME\n\nOBJECT = RANGEBEGINNINGDATE\nNUM_VAL = 1\nVALUE = \"{}\"\n" \
                           "END_OBJECT = RANGEBEGINNINGDATE\n\nOBJECT = RANGEBEGINNINGTIME\n" \
                           "NUM_VAL = 1\nVALUE = \"{}\"\n" \
                           "END_OBJECT = RANGEBEGINNINGTIME\n\nOBJECT = RANGEENDINGDATE\n" \
                           "NUM_VAL = 1\nVALUE = \"{}\"\n" \
                           "END_OBJECT = RANGEENDINGDATE\n\nOBJECT = RANGEENDINGTIME\nNUM_VAL = 1\nVALUE = \"{}\"\n" \
                           "END_OBJECT = RANGEENDINGTIME\nEND_GROUP = RANGEDATETIME"
    core_metadata_header = core_metadata_header.format(
        beginning_date.strftime("%Y-%m-%d"),
        beginning_date.strftime("%H:%M:%S.%f"),
        ending_date.strftime("%Y-%m-%d"),
        ending_date.strftime("%H:%M:%S.%f")
    )
    inst_metadata = "GROUP = ASSOCIATEDPLATFORMINSTRUMENTSENSOR\n\n" \
                    "OBJECT = ASSOCIATEDPLATFORMINSTRUMENTSENSORCONTAINER\nCLASS = \"1\"\n\n" \
                    "OBJECT = ASSOCIATEDSENSORSHORTNAME\nCLASS = \"1\"\nNUM_VAL = 1\n" \
                    "VALUE = \"MODIS\"\nEND_OBJECT = ASSOCIATEDSENSORSHORTNAME\n\n" \
                    "OBJECT = ASSOCIATEDPLATFORMSHORTNAME\nCLASS = \"1\"\nNUM_VAL = 1\n" \
                    "VALUE = \"Terra\"\nEND_OBJECT = ASSOCIATEDPLATFORMSHORTNAME\n\n" \
                    "OBJECT = ASSOCIATEDINSTRUMENTSHORTNAME\nCLASS = \"1\"\nNUM_VAL = 1\n" \
                    "VALUE = \"MODIS\"\nEND_OBJECT = ASSOCIATEDINSTRUMENTSHORTNAME\n\n" \
                    "END_OBJECT = ASSOCIATEDPLATFORMINSTRUMENTSENSORCONTAINER\n\n" \
                    "END_GROUP              = ASSOCIATEDPLATFORMINSTRUMENTSENSOR\n\n"
    collection_metadata = "GROUP = COLLECTIONDESCRIPTIONCLASS\n\nOBJECT = SHORTNAME\nNUM_VAL = 1\n" \
                          f"VALUE = \"{file_shortname}\"\nEND_OBJECT = SHORTNAME\n\n" \
                          "OBJECT = VERSIONID\nNUM_VAL = 1\nVALUE = 6\nEND_OBJECT = VERSIONID\n\n" \
                          "END_GROUP = COLLECTIONDESCRIPTIONCLASS\n\n"
    core_metadata_header += "\n\n" + inst_metadata + collection_metadata
    return core_metadata_header


def _create_struct_metadata(geo_resolution: int) -> str:
    geo_dim_factor = RES_TO_REPEAT_FACTOR[geo_resolution] * 2
    struct_metadata_header = "GROUP=SwathStructure\n" \
                             "GROUP=SWATH_1\n" \
                             "GROUP=DimensionMap\n" \
                             "OBJECT=DimensionMap_2\n" \
                             f"GeoDimension=\"{geo_dim_factor}*nscans\"\n" \
                             "END_OBJECT=DimensionMap_2\n" \
                             "END_GROUP=DimensionMap\n" \
                             "END_GROUP=SWATH_1\n" \
                             "END_GROUP=SwathStructure\nEND"
    return struct_metadata_header


def _create_header_metadata() -> str:
    archive_metadata_header = "GROUP = ARCHIVEDMETADATA\nEND_GROUP = ARCHIVEDMETADATA\nEND"
    return archive_metadata_header


@pytest.fixture(scope="session")
def modis_l1b_nasa_mod021km_file(tmpdir_factory) -> list[str]:
    """Create a single MOD021KM file following standard NASA file scheme."""
    filename = generate_nasa_l1b_filename("MOD021km")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=True)
    variable_infos.update(_get_visible_variable_info("EV_1KM_RefSB", 1000, AVAILABLE_1KM_VIS_PRODUCT_NAMES))
    variable_infos.update(_get_visible_variable_info("EV_500_Aggr1km_RefSB", 1000, AVAILABLE_HKM_PRODUCT_NAMES))
    variable_infos.update(_get_visible_variable_info("EV_250_Aggr1km_RefSB", 1000, AVAILABLE_QKM_PRODUCT_NAMES))
    variable_infos.update(_get_emissive_variable_info("EV_1KM_Emissive", 1000, AVAILABLE_1KM_IR_PRODUCT_NAMES))
    create_hdfeos_test_file(full_path, variable_infos, geo_resolution=5000, file_shortname="MOD021KM")
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_imapp_1000m_file(tmpdir_factory) -> list[str]:
    """Create a single MOD021KM file following IMAPP file scheme."""
    filename = generate_imapp_filename("1000m")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=True)
    variable_infos.update(_get_visible_variable_info("EV_1KM_RefSB", 1000, AVAILABLE_1KM_VIS_PRODUCT_NAMES))
    variable_infos.update(_get_visible_variable_info("EV_500_Aggr1km_RefSB", 1000, AVAILABLE_HKM_PRODUCT_NAMES))
    variable_infos.update(_get_visible_variable_info("EV_250_Aggr1km_RefSB", 1000, AVAILABLE_QKM_PRODUCT_NAMES))
    variable_infos.update(_get_emissive_variable_info("EV_1KM_Emissive", 1000, AVAILABLE_1KM_IR_PRODUCT_NAMES))
    create_hdfeos_test_file(full_path, variable_infos, geo_resolution=5000, file_shortname="MOD021KM")
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_nasa_mod02hkm_file(tmpdir_factory) -> list[str]:
    """Create a single MOD02HKM file following standard NASA file scheme."""
    filename = generate_nasa_l1b_filename("MOD02Hkm")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 1000, include_angles=False)
    variable_infos.update(_get_visible_variable_info("EV_500_RefSB", 250, AVAILABLE_QKM_PRODUCT_NAMES))
    create_hdfeos_test_file(full_path, variable_infos, geo_resolution=1000, file_shortname="MOD02HKM")
    return [full_path]


@pytest.fixture
def modis_l1b_nasa_mod02qkm_file(tmpdir_factory) -> list[str]:
    """Create a single MOD02QKM file following standard NASA file scheme."""
    filename = generate_nasa_l1b_filename("MOD02Qkm")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 1000, include_angles=False)
    variable_infos.update(_get_visible_variable_info("EV_250_RefSB", 250, AVAILABLE_QKM_PRODUCT_NAMES))
    create_hdfeos_test_file(full_path, variable_infos, geo_resolution=1000, file_shortname="MOD02QKM")
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_nasa_mod03_file(tmpdir_factory) -> list[str]:
    """Create a single MOD03 file following standard NASA file scheme."""
    filename = generate_nasa_l1b_filename("MOD03")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 1000, include_angles=True)
    create_hdfeos_test_file(full_path, variable_infos, geo_resolution=1000, file_shortname="MOD03")
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_imapp_geo_file(tmpdir_factory) -> list[str]:
    """Create a single geo file following standard IMAPP file scheme."""
    filename = generate_imapp_filename("geo")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 1000, include_angles=True)
    create_hdfeos_test_file(full_path, variable_infos, geo_resolution=1000, file_shortname="MOD03")
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_nasa_1km_mod03_files(modis_l1b_nasa_mod021km_file, modis_l1b_nasa_mod03_file) -> list[str]:
    """Create input files including the 1KM and MOD03 files."""
    return modis_l1b_nasa_mod021km_file + modis_l1b_nasa_mod03_file


# Level 2 Fixtures


def _get_basic_variable_info(var_name: str, resolution: int) -> dict:
    shape = _shape_for_resolution(resolution)
    data = np.zeros((shape[0], shape[1]), dtype=np.uint16)
    row_dim_name = f'Cell_Along_Swath_{resolution}m:modl2'
    col_dim_name = f'Cell_Across_Swath_{resolution}m:modl2'
    return {
        var_name: {
            'data': data,
            'type': SDC.UINT16,
            'fill_value': 0,
            'attrs': {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                'dim_labels': [row_dim_name,
                               col_dim_name],
                'valid_range': (0, 32767),
                'scale_factor': 1.,
                'add_offset': 0.,
            },
        },
    }


def _get_cloud_mask_variable_info(var_name: str, resolution: int) -> dict:
    num_bytes = 6
    shape = _shape_for_resolution(resolution)
    data = np.zeros((num_bytes, shape[0], shape[1]), dtype=np.int8)
    byte_dim_name = "Byte_Segment:mod35"
    row_dim_name = 'Cell_Along_Swath_1km:mod35'
    col_dim_name = 'Cell_Across_Swath_1km:mod35'
    return {
        var_name: {
            'data': data,
            'type': SDC.INT8,
            'fill_value': 0,
            'attrs': {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                'dim_labels': [byte_dim_name,
                               row_dim_name,
                               col_dim_name],
                'valid_range': (0, -1),
                'scale_factor': 1.,
                'add_offset': 0.,
            },
        },
        'Quality_Assurance': {
            'data': np.ones((shape[0], shape[1], 10), dtype=np.int8),
            'type': SDC.INT8,
            'fill_value': 0,
            'attrs': {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                'dim_labels': [row_dim_name,
                               col_dim_name,
                               'Quality_Dimension:mod35'],
                'valid_range': (0, -1),
                'scale_factor': 1.,
                'add_offset': 0.,
            },
        },
    }


def _get_mask_byte1_variable_info() -> dict:
    shape = _shape_for_resolution(1000)
    data = np.zeros((shape[0], shape[1]), dtype=np.uint16)
    row_dim_name = 'Cell_Along_Swath_1km:mod35'
    col_dim_name = 'Cell_Across_Swath_1km:mod35'
    return {
        "MODIS_Cloud_Mask": {
            'data': data,
            'type': SDC.UINT16,
            'fill_value': 9999,
            'attrs': {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                'dim_labels': [row_dim_name,
                               col_dim_name],
                'valid_range': (0, 4),
                'scale_factor': 1.,
                'add_offset': 0.,
            },

        },
        "MODIS_Simple_LandSea_Mask": {
            'data': data,
            'type': SDC.UINT16,
            'fill_value': 9999,
            'attrs': {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                'dim_labels': [row_dim_name,
                               col_dim_name],
                'valid_range': (0, 4),
                'scale_factor': 1.,
                'add_offset': 0.,
            },
        },
        "MODIS_Snow_Ice_Flag": {
            'data': data,
            'type': SDC.UINT16,
            'fill_value': 9999,
            'attrs': {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                'dim_labels': [row_dim_name,
                               col_dim_name],
                'valid_range': (0, 2),
                'scale_factor': 1.,
                'add_offset': 0.,
            },
        },
    }


def generate_nasa_l2_filename(prefix: str) -> str:
    """Generate a file name that follows MODIS 35 L2 convention in a temporary directory."""
    now = datetime.now()
    return f'{prefix}_L2.A{now:%Y%j.%H%M}.061.{now:%Y%j%H%M%S}.hdf'


@pytest.fixture(scope="session")
def modis_l2_nasa_mod35_file(tmpdir_factory) -> list[str]:
    """Create a single MOD35 L2 HDF4 file with headers."""
    filename = generate_nasa_l2_filename("MOD35")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=True)
    variable_infos.update(_get_cloud_mask_variable_info("Cloud_Mask", 1000))
    create_hdfeos_test_file(full_path, variable_infos, geo_resolution=5000, file_shortname="MOD35")
    return [full_path]


@pytest.fixture(scope="session")
def modis_l2_nasa_mod35_mod03_files(modis_l2_nasa_mod35_file, modis_l1b_nasa_mod03_file) -> list[str]:
    """Create a MOD35 L2 HDF4 file and MOD03 L1b geolocation file."""
    return modis_l2_nasa_mod35_file + modis_l1b_nasa_mod03_file


@pytest.fixture(scope="session")
def modis_l2_nasa_mod06_file(tmpdir_factory) -> list[str]:
    """Create a single MOD06 L2 HDF4 file with headers."""
    filename = generate_nasa_l2_filename("MOD06")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=True)
    variable_infos.update(_get_basic_variable_info("Surface_Pressure", 5000))
    create_hdfeos_test_file(full_path, variable_infos, geo_resolution=5000, file_shortname="MOD06")
    return [full_path]


@pytest.fixture(scope="session")
def modis_l2_imapp_snowmask_file(tmpdir_factory) -> list[str]:
    """Create a single IMAPP snowmask L2 HDF4 file with headers."""
    filename = generate_imapp_filename("snowmask")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=False)
    variable_infos.update(_get_basic_variable_info("Snow_Mask", 1000))
    create_hdfeos_test_file(full_path, variable_infos, include_metadata=False)
    return [full_path]


@pytest.fixture(scope="session")
def modis_l2_imapp_snowmask_geo_files(modis_l2_imapp_snowmask_file, modis_l1b_nasa_mod03_file) -> list[str]:
    """Create the IMAPP snowmask and geo HDF4 files."""
    return modis_l2_imapp_snowmask_file + modis_l1b_nasa_mod03_file


@pytest.fixture(scope="session")
def modis_l2_imapp_mask_byte1_file(tmpdir_factory) -> list[str]:
    """Create a single IMAPP mask_byte1 L2 HDF4 file with headers."""
    filename = generate_imapp_filename("mask_byte1")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=False)
    variable_infos.update(_get_mask_byte1_variable_info())
    create_hdfeos_test_file(full_path, variable_infos, include_metadata=False)
    return [full_path]


@pytest.fixture(scope="session")
def modis_l2_imapp_mask_byte1_geo_files(modis_l2_imapp_mask_byte1_file, modis_l1b_nasa_mod03_file) -> list[str]:
    """Create the IMAPP mask_byte1 and geo HDF4 files."""
    return modis_l2_imapp_mask_byte1_file + modis_l1b_nasa_mod03_file



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the hrpt reader."""

import os
import unittest
from contextlib import suppress
from tempfile import NamedTemporaryFile
from unittest import mock

import numpy as np
import xarray as xr

from satpy.readers.hrpt import HRPTFile, dtype
from satpy.tests.reader_tests.test_avhrr_l1b_gaclac import PygacPatcher
from satpy.tests.utils import make_dataid

NUMBER_OF_SCANS = 10
SWATH_WIDTH = 2048


class TestHRPTWithFile(unittest.TestCase):
    """Test base class with writing a fake file."""

    def setUp(self) -> None:
        """Set up the test case."""
        test_data = np.ones(NUMBER_OF_SCANS, dtype=dtype)
        # Channel 3a
        test_data["id"]["id"][:5] = 891
        # Channel 3b
        test_data["id"]["id"][5:] = 890
        with NamedTemporaryFile(mode='w+', suffix='.hmf', delete=False) as hrpt_file:
            self.filename = hrpt_file.name
            test_data.tofile(hrpt_file)

    def tearDown(self) -> None:
        """Tear down the test case."""
        with suppress(OSError):
            os.remove(self.filename)

    def _get_dataset(self, dataset_id):
        fh = HRPTFile(self.filename, {}, {})
        return fh.get_dataset(dataset_id, {})


class TestHRPTReading(TestHRPTWithFile):
    """Test case for reading hrpt data."""

    def test_reading(self):
        """Test that data is read."""
        fh = HRPTFile(self.filename, {}, {})
        assert fh._data is not None


class TestHRPTGetUncalibratedData(TestHRPTWithFile):
    """Test case for reading uncalibrated hrpt data."""

    def _get_channel_1_counts(self):
        return self._get_dataset(make_dataid(name='1', calibration='counts'))

    def test_get_dataset_returns_a_dataarray(self):
        """Test that get_dataset returns a dataarray."""
        result = self._get_channel_1_counts()
        assert isinstance(result, xr.DataArray)

    def test_platform_name(self):
        """Test that the platform name is correct."""
        result = self._get_channel_1_counts()
        assert result.attrs['platform_name'] == 'NOAA 19'

    def test_no_calibration_values_are_1(self):
        """Test that the values of non-calibrated data is 1."""
        result = self._get_channel_1_counts()
        assert (result.values == 1).all()


def fake_calibrate_solar(data, *args, **kwargs):
    """Fake calibration."""
    del args, kwargs
    return data * 25.43 + 3


def fake_calibrate_thermal(data, *args, **kwargs):
    """Fake calibration."""
    del args, kwargs
    return data * 35.43 + 3


class CalibratorPatcher(PygacPatcher):
    """Patch pygac."""

    def setUp(self) -> None:
        """Patch pygac's calibration."""
        super().setUp()

        # Import things to patch here to make them patchable. Otherwise another function
        # might import it first which would prevent a successful patch.
        from pygac.calibration import Calibrator, calibrate_solar, calibrate_thermal
        self.Calibrator = Calibrator
        self.calibrate_thermal = calibrate_thermal
        self.calibrate_thermal.side_effect = fake_calibrate_thermal
        self.calibrate_solar = calibrate_solar
        self.calibrate_solar.side_effect = fake_calibrate_solar


class TestHRPTWithPatchedCalibratorAndFile(CalibratorPatcher, TestHRPTWithFile):
    """Test case with patched calibration routines and a synthetic file."""

    def setUp(self) -> None:
        """Set up the test case."""
        CalibratorPatcher.setUp(self)
        TestHRPTWithFile.setUp(self)

    def tearDown(self):
        """Tear down the test case."""
        CalibratorPatcher.tearDown(self)
        TestHRPTWithFile.tearDown(self)


class TestHRPTGetCalibratedReflectances(TestHRPTWithPatchedCalibratorAndFile):
    """Test case for reading calibrated reflectances from hrpt data."""

    def _get_channel_1_reflectance(self):
        """Get the channel 1 reflectance."""
        dataset_id = make_dataid(name='1', calibration='reflectance')
        return self._get_dataset(dataset_id)

    def test_calibrated_reflectances_values(self):
        """Test the calibrated reflectance values."""
        result = self._get_channel_1_reflectance()
        np.testing.assert_allclose(result.values, 28.43)


class TestHRPTGetCalibratedBT(TestHRPTWithPatchedCalibratorAndFile):
    """Test case for reading calibrated brightness temperature from hrpt data."""

    def _get_channel_4_bt(self):
        """Get the channel 4 bt."""
        dataset_id = make_dataid(name='4', calibration='brightness_temperature')
        return self._get_dataset(dataset_id)

    def test_calibrated_bt_values(self):
        """Test the calibrated reflectance values."""
        result = self._get_channel_4_bt()
        np.testing.assert_allclose(result.values, 38.43)


class TestHRPTChannel3(TestHRPTWithPatchedCalibratorAndFile):
    """Test case for reading calibrated brightness temperature from hrpt data."""

    def _get_channel_3b_bt(self):
        """Get the channel 4 bt."""
        dataset_id = make_dataid(name='3b', calibration='brightness_temperature')
        return self._get_dataset(dataset_id)

    def _get_channel_3a_reflectance(self):
        """Get the channel 4 bt."""
        dataset_id = make_dataid(name='3a', calibration='reflectance')
        return self._get_dataset(dataset_id)

    def _get_channel_3a_counts(self):
        """Get the channel 4 bt."""
        dataset_id = make_dataid(name='3a', calibration='counts')
        return self._get_dataset(dataset_id)

    def test_channel_3b_masking(self):
        """Test that channel 3b is split correctly."""
        result = self._get_channel_3b_bt()
        assert np.isnan(result.values[:5]).all()
        assert np.isfinite(result.values[5:]).all()

    def test_channel_3a_masking(self):
        """Test that channel 3a is split correctly."""
        result = self._get_channel_3a_reflectance()
        assert np.isnan(result.values[5:]).all()
        assert np.isfinite(result.values[:5]).all()

    def test_uncalibrated_channel_3a_masking(self):
        """Test that channel 3a is split correctly."""
        result = self._get_channel_3a_counts()
        assert np.isnan(result.values[5:]).all()
        assert np.isfinite(result.values[:5]).all()


class TestHRPTNavigation(TestHRPTWithFile):
    """Test case for computing HRPT navigation."""

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()
        self.fake_lons = np.ones((NUMBER_OF_SCANS, SWATH_WIDTH))
        self.fake_lats = np.ones((NUMBER_OF_SCANS, SWATH_WIDTH)) * 2

    def _prepare_mocks(self, Orbital, SatelliteInterpolator, get_lonlatalt):
        """Prepare the mocks."""
        Orbital.return_value.get_position.return_value = mock.MagicMock(), mock.MagicMock()
        get_lonlatalt.return_value = (mock.MagicMock(), mock.MagicMock(), mock.MagicMock())
        SatelliteInterpolator.return_value.interpolate.return_value = self.fake_lons, self.fake_lats

    @mock.patch.multiple('satpy.readers.hrpt',
                         Orbital=mock.DEFAULT,
                         compute_pixels=mock.DEFAULT,
                         get_lonlatalt=mock.DEFAULT,
                         SatelliteInterpolator=mock.DEFAULT)
    def test_longitudes_are_returned(self, Orbital, compute_pixels, get_lonlatalt, SatelliteInterpolator):
        """Check that latitudes are returned properly."""
        self._prepare_mocks(Orbital, SatelliteInterpolator, get_lonlatalt)
        dataset_id = make_dataid(name='longitude')
        result = self._get_dataset(dataset_id)
        assert (result == self.fake_lons).all()

    @mock.patch.multiple('satpy.readers.hrpt',
                         Orbital=mock.DEFAULT,
                         compute_pixels=mock.DEFAULT,
                         get_lonlatalt=mock.DEFAULT,
                         SatelliteInterpolator=mock.DEFAULT)
    def test_latitudes_are_returned(self, Orbital, compute_pixels, get_lonlatalt, SatelliteInterpolator):
        """Check that latitudes are returned properly."""
        self._prepare_mocks(Orbital, SatelliteInterpolator, get_lonlatalt)
        dataset_id = make_dataid(name='latitude')
        result = self._get_dataset(dataset_id)
        assert (result == self.fake_lats).all()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Writer utilities."""


def flatten_dict(d, parent_key='', sep='_'):
    """Flatten a nested dictionary.

    Based on https://stackoverflow.com/a/6027615/5703449
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, parent_key=new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test the MSG common (native and hrit format) functionionalities."""

import unittest
from datetime import datetime

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy import CHUNK_SIZE
from satpy.readers.seviri_base import (
    NoValidOrbitParams,
    OrbitPolynomial,
    OrbitPolynomialFinder,
    chebyshev,
    dec10216,
    get_cds_time,
    get_padding_area,
    get_satpos,
    pad_data_horizontally,
    pad_data_vertically,
)


def chebyshev4(c, x, domain):
    """Evaluate 4th order Chebyshev polynomial."""
    start_x, end_x = domain
    t = (x - 0.5 * (end_x + start_x)) / (0.5 * (end_x - start_x))
    return c[0] + c[1]*t + c[2]*(2*t**2 - 1) + c[3]*(4*t**3 - 3*t) - 0.5*c[0]


class SeviriBaseTest(unittest.TestCase):
    """Test SEVIRI base."""

    def test_dec10216(self):
        """Test the dec10216 function."""
        res = dec10216(np.array([255, 255, 255, 255, 255], dtype=np.uint8))
        exp = (np.ones((4, )) * 1023).astype(np.uint16)
        np.testing.assert_equal(res, exp)
        res = dec10216(np.array([1, 1, 1, 1, 1], dtype=np.uint8))
        exp = np.array([4,  16,  64, 257], dtype=np.uint16)
        np.testing.assert_equal(res, exp)

    def test_chebyshev(self):
        """Test the chebyshev function."""
        coefs = [1, 2, 3, 4]
        time = 123
        domain = [120, 130]
        res = chebyshev(coefs=[1, 2, 3, 4], time=time, domain=domain)
        exp = chebyshev4(coefs, time, domain)
        np.testing.assert_allclose(res, exp)

    def test_get_cds_time(self):
        """Test the get_cds_time function."""
        # Scalar
        self.assertEqual(get_cds_time(days=21246, msecs=12*3600*1000),
                         np.datetime64('2016-03-03 12:00'))

        # Array
        days = np.array([21246, 21247, 21248])
        msecs = np.array([12*3600*1000, 13*3600*1000 + 1, 14*3600*1000 + 2])
        expected = np.array([np.datetime64('2016-03-03 12:00:00.000'),
                             np.datetime64('2016-03-04 13:00:00.001'),
                             np.datetime64('2016-03-05 14:00:00.002')])
        np.testing.assert_equal(get_cds_time(days=days, msecs=msecs), expected)

        days = 21246
        msecs = 12*3600*1000
        expected = np.datetime64('2016-03-03 12:00:00.000')
        np.testing.assert_equal(get_cds_time(days=days, msecs=msecs), expected)

    def test_pad_data_horizontally_bad_shape(self):
        """Test the error handling for the horizontal hrv padding."""
        data = xr.DataArray(data=np.zeros((1, 10)), dims=('y', 'x'))
        east_bound = 5
        west_bound = 10
        final_size = (1, 20)
        with self.assertRaises(IndexError):
            pad_data_horizontally(data, final_size, east_bound, west_bound)

    def test_pad_data_vertically_bad_shape(self):
        """Test the error handling for the vertical hrv padding."""
        data = xr.DataArray(data=np.zeros((10, 1)), dims=('y', 'x'))
        south_bound = 5
        north_bound = 10
        final_size = (20, 1)
        with self.assertRaises(IndexError):
            pad_data_vertically(data, final_size, south_bound, north_bound)

    @staticmethod
    def test_pad_data_horizontally():
        """Test the horizontal hrv padding."""
        data = xr.DataArray(data=np.zeros((1, 10)), dims=('y', 'x'))
        east_bound = 4
        west_bound = 13
        final_size = (1, 20)
        res = pad_data_horizontally(data, final_size, east_bound, west_bound)
        expected = np.array([[np.nan, np.nan, np.nan,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
                              np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]])
        np.testing.assert_equal(res, expected)

    @staticmethod
    def test_pad_data_vertically():
        """Test the vertical hrv padding."""
        data = xr.DataArray(data=np.zeros((10, 1)), dims=('y', 'x'))
        south_bound = 4
        north_bound = 13
        final_size = (20, 1)
        res = pad_data_vertically(data, final_size, south_bound, north_bound)
        expected = np.zeros(final_size)
        expected[:] = np.nan
        expected[south_bound-1:north_bound] = 0.
        np.testing.assert_equal(res, expected)

    @staticmethod
    def test_get_padding_area_float():
        """Test padding area generator for floats."""
        shape = (10, 10)
        dtype = np.float64
        res = get_padding_area(shape, dtype)
        expected = da.full(shape, np.nan, dtype=dtype, chunks=CHUNK_SIZE)
        np.testing.assert_array_equal(res, expected)

    @staticmethod
    def test_get_padding_area_int():
        """Test padding area generator for integers."""
        shape = (10, 10)
        dtype = np.int64
        res = get_padding_area(shape, dtype)
        expected = da.full(shape, 0, dtype=dtype, chunks=CHUNK_SIZE)
        np.testing.assert_array_equal(res, expected)


ORBIT_POLYNOMIALS = {
    'StartTime': np.array([
        [
            datetime(2006, 1, 1, 6), datetime(2006, 1, 1, 12),
            datetime(2006, 1, 1, 18), datetime(1958, 1, 1, 0)]
    ]),
    'EndTime': np.array([
        [
            datetime(2006, 1, 1, 12), datetime(2006, 1, 1, 18),
            datetime(2006, 1, 2, 0), datetime(1958, 1, 1, 0)
        ]
    ]),
    'X': [np.zeros(8),
          [8.41607082e+04, 2.94319260e+00, 9.86748617e-01,
           -2.70135453e-01,
           -3.84364650e-02, 8.48718433e-03, 7.70548174e-04,
           -1.44262718e-04],
          np.zeros(8)],
    'Y': [np.zeros(8),
          [-5.21170255e+03, 5.12998948e+00, -1.33370453e+00,
           -3.09634144e-01,
           6.18232793e-02, 7.50505681e-03, -1.35131011e-03,
           -1.12054405e-04],
          np.zeros(8)],
    'Z': [np.zeros(8),
          [-6.51293855e+02, 1.45830459e+02, 5.61379400e+01,
           -3.90970565e+00,
           -7.38137565e-01, 3.06131644e-02, 3.82892428e-03,
           -1.12739309e-04],
          np.zeros(8)],
}
ORBIT_POLYNOMIALS_SYNTH = {
    # 12-31: Contiguous
    # 01-01: Small gap (12:00 - 13:00)
    # 01-02: Large gap (04:00 - 18:00)
    # 01-03: Overlap (10:00 - 13:00)
    'StartTime': np.array([
        [
            datetime(2005, 12, 31, 10), datetime(2005, 12, 31, 12),
            datetime(2006, 1, 1, 10), datetime(2006, 1, 1, 13),
            datetime(2006, 1, 2, 0), datetime(2006, 1, 2, 18),
            datetime(2006, 1, 3, 6), datetime(2006, 1, 3, 10),
        ]
    ]),
    'EndTime': np.array([
        [
            datetime(2005, 12, 31, 12), datetime(2005, 12, 31, 18),
            datetime(2006, 1, 1, 12), datetime(2006, 1, 1, 18),
            datetime(2006, 1, 2, 4), datetime(2006, 1, 2, 22),
            datetime(2006, 1, 3, 13), datetime(2006, 1, 3, 18),
        ]
    ]),
    'X': [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0],
    'Y': [1.1, 2.1, 3.1, 4.1, 5.1, 6.1, 7.1, 8.1],
    'Z': [1.2, 2.2, 3.2, 4.2, 5.2, 6.2, 7.2, 8.2],
}
ORBIT_POLYNOMIALS_INVALID = {
    'StartTime': np.array([
        [
            datetime(1958, 1, 1), datetime(1958, 1, 1)
        ]
    ]),
    'EndTime': np.array([
        [
            datetime(1958, 1, 1), datetime(1958, 1, 1)
        ]
    ]),
    'X': [1, 2],
    'Y': [3, 4],
    'Z': [5, 6],
}


class TestSatellitePosition:
    """Test locating the satellite."""

    @pytest.fixture
    def orbit_polynomial(self):
        """Get an orbit polynomial for testing."""
        return OrbitPolynomial(
            start_time=datetime(2006, 1, 1, 12),
            end_time=datetime(2006, 1, 1, 18),
            coefs=(
                np.array([8.41607082e+04, 2.94319260e+00, 9.86748617e-01,
                          -2.70135453e-01, -3.84364650e-02, 8.48718433e-03,
                          7.70548174e-04, -1.44262718e-04]),
                np.array([-5.21170255e+03, 5.12998948e+00, -1.33370453e+00,
                          -3.09634144e-01, 6.18232793e-02, 7.50505681e-03,
                          -1.35131011e-03, -1.12054405e-04]),
                np.array([-6.51293855e+02, 1.45830459e+02, 5.61379400e+01,
                          -3.90970565e+00, -7.38137565e-01, 3.06131644e-02,
                          3.82892428e-03, -1.12739309e-04])
            )
        )

    @pytest.fixture
    def time(self):
        """Get scan timestamp for testing."""
        return datetime(2006, 1, 1, 12, 15, 9, 304888)

    def test_eval_polynomial(self, orbit_polynomial, time):
        """Test getting the position in cartesian coordinates."""
        x, y, z = orbit_polynomial.evaluate(time)
        np.testing.assert_allclose(
            [x, y, z],
            [42078421.37095518, -2611352.744615312, -419828.9699940758]
        )

    def test_get_satpos(self, orbit_polynomial, time):
        """Test getting the position in geodetic coordinates."""
        lon, lat, alt = get_satpos(
            orbit_polynomial=orbit_polynomial,
            time=time,
            semi_major_axis=6378169.00,
            semi_minor_axis=6356583.80
        )
        np.testing.assert_allclose(
            [lon, lat, alt],
            [-3.55117540817073, -0.5711243456528018, 35783296.150123544]
        )


class TestOrbitPolynomialFinder:
    """Unit tests for orbit polynomial finder."""

    @pytest.mark.parametrize(
        ('orbit_polynomials', 'time', 'orbit_polynomial_exp'),
        [
            # Contiguous validity intervals (that's the norm)
            (
                ORBIT_POLYNOMIALS_SYNTH,
                datetime(2005, 12, 31, 12, 15),
                OrbitPolynomial(
                    coefs=(2.0, 2.1, 2.2),
                    start_time=np.datetime64('2005-12-31 12:00'),
                    end_time=np.datetime64('2005-12-31 18:00')
                )
            ),
            # No interval enclosing the given timestamp, but closest interval
            # not too far away
            (
                    ORBIT_POLYNOMIALS_SYNTH,
                    datetime(2006, 1, 1, 12, 15),
                    OrbitPolynomial(
                        coefs=(3.0, 3.1, 3.2),
                        start_time=np.datetime64('2006-01-01 10:00'),
                        end_time=np.datetime64('2006-01-01 12:00')
                    )
            ),
            # Overlapping intervals
            (
                    ORBIT_POLYNOMIALS_SYNTH,
                    datetime(2006, 1, 3, 12, 15),
                    OrbitPolynomial(
                        coefs=(8.0, 8.1, 8.2),
                        start_time=np.datetime64('2006-01-03 10:00'),
                        end_time=np.datetime64('2006-01-03 18:00')
                    )
            ),
        ]
    )
    def test_get_orbit_polynomial(self, orbit_polynomials, time,
                                  orbit_polynomial_exp):
        """Test getting the satellite locator."""
        finder = OrbitPolynomialFinder(orbit_polynomials)
        orbit_polynomial = finder.get_orbit_polynomial(time=time)
        assert orbit_polynomial == orbit_polynomial_exp

    @pytest.mark.parametrize(
        ('orbit_polynomials', 'time'),
        [
            # No interval enclosing the given timestamp and closest interval
            # too far away
            (ORBIT_POLYNOMIALS_SYNTH, datetime(2006, 1, 2, 12, 15)),
            # No valid polynomials at all
            (ORBIT_POLYNOMIALS_INVALID, datetime(2006, 1, 1, 12, 15))
        ]
    )
    def test_get_orbit_polynomial_exceptions(self, orbit_polynomials, time):
        """Test exceptions thrown while getting the satellite locator."""
        finder = OrbitPolynomialFinder(orbit_polynomials)
        with pytest.raises(NoValidOrbitParams):
            finder.get_orbit_polynomial(time=time)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The ahi_hsd reader tests package."""

import unittest
import warnings
from datetime import datetime
from unittest import mock

import dask.array as da
import numpy as np

from satpy.readers.ahi_hsd import AHIHSDFileHandler
from satpy.readers.utils import get_geostationary_mask


class TestAHIHSDNavigation(unittest.TestCase):
    """Test the AHI HSD reader navigation."""

    @mock.patch('satpy.readers.ahi_hsd.np2str')
    @mock.patch('satpy.readers.ahi_hsd.np.fromfile')
    def test_region(self, fromfile, np2str):
        """Test region navigation."""
        from pyresample.utils import proj4_radius_parameters
        np2str.side_effect = lambda x: x
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_hsd.open', m, create=True):
            fh = AHIHSDFileHandler('somefile',
                                   {'segment': 1, 'total_segments': 1},
                                   filetype_info={'file_type': 'hsd_b01'},
                                   user_calibration=None)
            fh.proj_info = {'CFAC': 40932549,
                            'COFF': -591.5,
                            'LFAC': 40932549,
                            'LOFF': 5132.5,
                            'blocklength': 127,
                            'coeff_for_sd': 1737122264.0,
                            'distance_from_earth_center': 42164.0,
                            'earth_equatorial_radius': 6378.137,
                            'earth_polar_radius': 6356.7523,
                            'hblock_number': 3,
                            'req2_rpol2': 1.006739501,
                            'req2_rpol2_req2': 0.0066943844,
                            'resampling_size': 4,
                            'resampling_types': 0,
                            'rpol2_req2': 0.993305616,
                            'spare': '',
                            'sub_lon': 140.7}

            fh.data_info = {'blocklength': 50,
                            'compression_flag_for_data': 0,
                            'hblock_number': 2,
                            'number_of_bits_per_pixel': 16,
                            'number_of_columns': 1000,
                            'number_of_lines': 1000,
                            'spare': ''}

            area_def = fh.get_area_def(None)
            proj_dict = area_def.proj_dict
            a, b = proj4_radius_parameters(proj_dict)
            self.assertEqual(a, 6378137.0)
            self.assertEqual(b, 6356752.3)
            self.assertEqual(proj_dict['h'], 35785863.0)
            self.assertEqual(proj_dict['lon_0'], 140.7)
            self.assertEqual(proj_dict['proj'], 'geos')
            self.assertEqual(proj_dict['units'], 'm')
            np.testing.assert_allclose(area_def.area_extent, (592000.0038256242, 4132000.0267018233,
                                                              1592000.0102878273, 5132000.033164027))

    @mock.patch('satpy.readers.ahi_hsd.np2str')
    @mock.patch('satpy.readers.ahi_hsd.np.fromfile')
    def test_segment(self, fromfile, np2str):
        """Test segment navigation."""
        from pyresample.utils import proj4_radius_parameters
        np2str.side_effect = lambda x: x
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_hsd.open', m, create=True):
            fh = AHIHSDFileHandler('somefile', {'segment': 8, 'total_segments': 10},
                                   filetype_info={'file_type': 'hsd_b01'})
            fh.proj_info = {'CFAC': 40932549,
                            'COFF': 5500.5,
                            'LFAC': 40932549,
                            'LOFF': 5500.5,
                            'blocklength': 127,
                            'coeff_for_sd': 1737122264.0,
                            'distance_from_earth_center': 42164.0,
                            'earth_equatorial_radius': 6378.137,
                            'earth_polar_radius': 6356.7523,
                            'hblock_number': 3,
                            'req2_rpol2': 1.006739501,
                            'req2_rpol2_req2': 0.0066943844,
                            'resampling_size': 4,
                            'resampling_types': 0,
                            'rpol2_req2': 0.993305616,
                            'spare': '',
                            'sub_lon': 140.7}

            fh.data_info = {'blocklength': 50,
                            'compression_flag_for_data': 0,
                            'hblock_number': 2,
                            'number_of_bits_per_pixel': 16,
                            'number_of_columns': 11000,
                            'number_of_lines': 1100,
                            'spare': ''}

            area_def = fh.get_area_def(None)
            proj_dict = area_def.proj_dict
            a, b = proj4_radius_parameters(proj_dict)
            self.assertEqual(a, 6378137.0)
            self.assertEqual(b, 6356752.3)
            self.assertEqual(proj_dict['h'], 35785863.0)
            self.assertEqual(proj_dict['lon_0'], 140.7)
            self.assertEqual(proj_dict['proj'], 'geos')
            self.assertEqual(proj_dict['units'], 'm')
            np.testing.assert_allclose(area_def.area_extent, (-5500000.035542117, -3300000.021325271,
                                                              5500000.035542117, -2200000.0142168473))


class TestAHIHSDFileHandler(unittest.TestCase):
    """Test case for the file reading."""

    def new_unzip(fname):
        """Fake unzipping."""
        if fname[-3:] == 'bz2':
            return fname[:-4]
        return fname

    @staticmethod
    def _create_fake_file_handler(in_fname, filename_info=None, filetype_info=None):
        if filename_info is None:
            filename_info = {'segment': 8, 'total_segments': 10}
        if filetype_info is None:
            filetype_info = {'file_type': 'hsd_b01'}
        fh = AHIHSDFileHandler(in_fname, filename_info, filetype_info)

        # Check that the filename is altered for bz2 format files
        assert in_fname != fh.filename

        fh.proj_info = {
            'CFAC': 40932549,
            'COFF': 5500.5,
            'LFAC': 40932549,
            'LOFF': 5500.5,
            'blocklength': 127,
            'coeff_for_sd': 1737122264.0,
            'distance_from_earth_center': 42164.0,
            'earth_equatorial_radius': 6378.137,
            'earth_polar_radius': 6356.7523,
            'hblock_number': 3,
            'req2_rpol2': 1.006739501,
            'req2_rpol2_req2': 0.0066943844,
            'resampling_size': 4,
            'resampling_types': 0,
            'rpol2_req2': 0.993305616,
            'spare': '',
            'sub_lon': 140.7
        }
        fh.nav_info = {
            'SSP_longitude': 140.66,
            'SSP_latitude': 0.03,
            'distance_earth_center_to_satellite': 42165.04,
            'nadir_longitude': 140.67,
            'nadir_latitude': 0.04
        }
        fh.data_info = {
            'blocklength': 50,
            'compression_flag_for_data': 0,
            'hblock_number': 2,
            'number_of_bits_per_pixel': 16,
            'number_of_columns': 11000,
            'number_of_lines': 1100,
            'spare': ''
        }
        fh.basic_info = {
            'observation_area': np.array(['FLDK']),
            'observation_start_time': np.array([58413.12523839]),
            'observation_end_time': np.array([58413.12562439]),
            'observation_timeline': np.array([300]),
        }
        fh.observation_area = fh.basic_info['observation_area']
        return fh

    @mock.patch('satpy.readers.ahi_hsd.np2str')
    @mock.patch('satpy.readers.ahi_hsd.np.fromfile')
    @mock.patch('satpy.readers.ahi_hsd.unzip_file',
                mock.MagicMock(side_effect=new_unzip))
    def setUp(self, fromfile, np2str):
        """Create a test file handler."""
        np2str.side_effect = lambda x: x
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_hsd.open', m, create=True):
            # Check if file handler raises exception for invalid calibration mode
            with self.assertRaises(ValueError):
                AHIHSDFileHandler('somefile',
                                  {'segment': 8, 'total_segments': 10},
                                  filetype_info={'file_type': 'hsd_b01'},
                                  calib_mode='BAD_MODE')
            in_fname = 'test_file.bz2'
            self.fh = self._create_fake_file_handler(in_fname)

    def test_time_properties(self):
        """Test start/end/scheduled time properties."""
        self.assertEqual(self.fh.start_time, datetime(2018, 10, 22, 3, 0, 20, 596896))
        self.assertEqual(self.fh.end_time, datetime(2018, 10, 22, 3, 0, 53, 947296))
        self.assertEqual(self.fh.scheduled_time, datetime(2018, 10, 22, 3, 0, 0, 0))

    def test_scanning_frequencies(self):
        """Test scanning frequencies."""
        self.fh.observation_area = 'JP04'
        self.assertEqual(self.fh.scheduled_time, datetime(2018, 10, 22, 3, 7, 30, 0))
        self.fh.observation_area = 'R304'
        self.assertEqual(self.fh.scheduled_time, datetime(2018, 10, 22, 3, 7, 30, 0))
        self.fh.observation_area = 'R420'
        self.assertEqual(self.fh.scheduled_time, datetime(2018, 10, 22, 3, 9, 30, 0))
        self.fh.observation_area = 'R520'
        self.assertEqual(self.fh.scheduled_time, datetime(2018, 10, 22, 3, 9, 30, 0))
        self.fh.observation_area = 'FLDK'
        self.assertEqual(self.fh.scheduled_time, datetime(2018, 10, 22, 3, 0, 0, 0))

    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._read_header')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._read_data')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._mask_invalid')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler.calibrate')
    def test_read_band(self, calibrate, *mocks):
        """Test masking of space pixels."""
        nrows = 25
        ncols = 100
        self.fh.data_info['number_of_columns'] = ncols
        self.fh.data_info['number_of_lines'] = nrows
        calibrate.return_value = np.ones((nrows, ncols))
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_hsd.open', m, create=True):
            im = self.fh.read_band(info=mock.MagicMock(), key=mock.MagicMock())
            # Note: Within the earth's shape get_geostationary_mask() is True but the numpy.ma mask
            # is False
            mask = im.to_masked_array().mask
            ref_mask = np.logical_not(get_geostationary_mask(self.fh.area).compute())
            self.assertTrue(np.all(mask == ref_mask))

            # Test attributes
            orb_params_exp = {'projection_longitude': 140.7,
                              'projection_latitude': 0.,
                              'projection_altitude': 35785863.0,
                              'satellite_actual_longitude': 140.66,
                              'satellite_actual_latitude': 0.03,
                              'nadir_longitude': 140.67,
                              'nadir_latitude': 0.04}
            self.assertTrue(set(orb_params_exp.items()).issubset(set(im.attrs['orbital_parameters'].items())))
            self.assertTrue(np.isclose(im.attrs['orbital_parameters']['satellite_actual_altitude'], 35786903.00581372))

            # Test if masking space pixels disables with appropriate flag
            self.fh.mask_space = False
            with mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._mask_space') as mask_space:
                self.fh.read_band(info=mock.MagicMock(), key=mock.MagicMock())
                mask_space.assert_not_called()

    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._read_header')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._read_data')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._mask_invalid')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler.calibrate')
    def test_scene_loading(self, calibrate, *mocks):
        """Test masking of space pixels."""
        from satpy import Scene
        nrows = 25
        ncols = 100
        calibrate.return_value = np.ones((nrows, ncols))
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_hsd.open', m, create=True), \
             mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler') as fh_cls:
            fh_cls.return_value = self.fh
            self.fh.filename_info['total_segments'] = 1
            self.fh.filename_info['segment'] = 1
            self.fh.data_info['number_of_columns'] = ncols
            self.fh.data_info['number_of_lines'] = nrows
            scn = Scene(reader='ahi_hsd', filenames=['HS_H08_20210225_0700_B07_FLDK_R20_S0110.DAT'])
            scn.load(['B07'])
            im = scn['B07']

            # Make sure space masking worked
            mask = im.to_masked_array().mask
            ref_mask = np.logical_not(get_geostationary_mask(self.fh.area).compute())
            self.assertTrue(np.all(mask == ref_mask))

    def test_blocklen_error(self, *mocks):
        """Test erraneous blocklength."""
        open_name = '%s.open' % __name__
        fpos = 50
        with mock.patch(open_name, create=True) as mock_open:
            with mock_open(mock.MagicMock(), 'r') as fp_:
                # Expected and actual blocklength match
                fp_.tell.return_value = 50
                with warnings.catch_warnings(record=True) as w:
                    self.fh._check_fpos(fp_, fpos, 0, 'header 1')
                    self.assertTrue(len(w) == 0)

                # Expected and actual blocklength do not match
                fp_.tell.return_value = 100
                with warnings.catch_warnings(record=True) as w:
                    self.fh._check_fpos(fp_, fpos, 0, 'header 1')
                    self.assertTrue(len(w) > 0)

    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._check_fpos')
    def test_read_header(self, *mocks):
        """Test header reading."""
        nhdr = [
            {'blocklength': 0},
            {'blocklength': 0},
            {'blocklength': 0},
            {'blocklength': 0},
            {'blocklength': 0, 'band_number': [4]},
            {'blocklength': 0},
            {'blocklength': 0},
            {'blocklength': 0},
            {'blocklength': 0, 'numof_correction_info_data': [1]},
            {'blocklength': 0},
            {'blocklength': 0, 'number_of_observation_times': [1]},
            {'blocklength': 0},
            {'blocklength': 0, 'number_of_error_info_data': [1]},
            {'blocklength': 0},
            {'blocklength': 0}]
        with mock.patch('numpy.fromfile', side_effect=nhdr):
            self.fh._read_header(mock.MagicMock())


class TestAHICalibration(unittest.TestCase):
    """Test case for various AHI calibration types."""

    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler.__init__',
                return_value=None)
    def setUp(self, *mocks):
        """Create fake data for testing."""
        self.def_cali = [-0.0037, 15.20]
        self.upd_cali = [-0.0074, 30.40]
        self.bad_cali = [0.0, 0.0]
        fh = AHIHSDFileHandler(filetype_info={'file_type': 'hsd_b01'})
        fh.calib_mode = 'NOMINAL'
        fh.user_calibration = None
        fh.is_zipped = False
        fh._header = {
            'block5': {'band_number': [5],
                       'gain_count2rad_conversion': [self.def_cali[0]],
                       'offset_count2rad_conversion': [self.def_cali[1]],
                       'central_wave_length': [10.4073], },
            'calibration': {'coeff_rad2albedo_conversion': [0.0019255],
                            'speed_of_light': [299792458.0],
                            'planck_constant': [6.62606957e-34],
                            'boltzmann_constant': [1.3806488e-23],
                            'c0_rad2tb_conversion': [-0.116127314574],
                            'c1_rad2tb_conversion': [1.00099153832],
                            'c2_rad2tb_conversion': [-1.76961091571e-06],
                            'cali_gain_count2rad_conversion': [self.upd_cali[0]],
                            'cali_offset_count2rad_conversion': [self.upd_cali[1]]},
        }

        self.counts = da.array(np.array([[0., 1000.],
                                         [2000., 5000.]]))
        self.fh = fh

    def test_default_calibrate(self, *mocks):
        """Test default in-file calibration modes."""
        self.setUp()
        # Counts
        self.assertEqual(self.fh.calibrate(data=123,
                                           calibration='counts'),
                         123)

        # Radiance
        rad_exp = np.array([[15.2, 11.5],
                            [7.8, -3.3]])
        rad = self.fh.calibrate(data=self.counts,
                                calibration='radiance')
        self.assertTrue(np.allclose(rad, rad_exp))

        # Brightness Temperature
        bt_exp = np.array([[330.978979, 310.524688],
                           [285.845017, np.nan]])
        bt = self.fh.calibrate(data=self.counts,
                               calibration='brightness_temperature')
        np.testing.assert_allclose(bt, bt_exp)

        # Reflectance
        refl_exp = np.array([[2.92676, 2.214325],
                             [1.50189, 0.]])
        refl = self.fh.calibrate(data=self.counts,
                                 calibration='reflectance')
        self.assertTrue(np.allclose(refl, refl_exp))

    def test_updated_calibrate(self):
        """Test updated in-file calibration modes."""
        # Standard operation
        self.fh.calib_mode = 'UPDATE'
        rad_exp = np.array([[30.4, 23.0],
                            [15.6, -6.6]])
        rad = self.fh.calibrate(data=self.counts, calibration='radiance')
        self.assertTrue(np.allclose(rad, rad_exp))

        # Case for no updated calibration available (older data)
        self.fh._header = {
            'block5': {'band_number': [5],
                       'gain_count2rad_conversion': [self.def_cali[0]],
                       'offset_count2rad_conversion': [self.def_cali[1]],
                       'central_wave_length': [10.4073], },
            'calibration': {'coeff_rad2albedo_conversion': [0.0019255],
                            'speed_of_light': [299792458.0],
                            'planck_constant': [6.62606957e-34],
                            'boltzmann_constant': [1.3806488e-23],
                            'c0_rad2tb_conversion': [-0.116127314574],
                            'c1_rad2tb_conversion': [1.00099153832],
                            'c2_rad2tb_conversion': [-1.76961091571e-06],
                            'cali_gain_count2rad_conversion': [self.bad_cali[0]],
                            'cali_offset_count2rad_conversion': [self.bad_cali[1]]},
        }
        rad = self.fh.calibrate(data=self.counts, calibration='radiance')
        rad_exp = np.array([[15.2, 11.5],
                            [7.8, -3.3]])
        self.assertTrue(np.allclose(rad, rad_exp))

    def test_user_calibration(self):
        """Test user-defined calibration modes."""
        # This is for radiance correction
        self.fh.user_calibration = {'B13': {'slope': 0.95,
                                            'offset': -0.1}}
        self.fh.band_name = 'B13'
        rad = self.fh.calibrate(data=self.counts, calibration='radiance').compute()
        rad_exp = np.array([[16.10526316, 12.21052632],
                            [8.31578947, -3.36842105]])
        self.assertTrue(np.allclose(rad, rad_exp))

        # This is for DN calibration
        self.fh.user_calibration = {'B13': {'slope': -0.0032,
                                            'offset': 15.20},
                                    'type': 'DN'}
        self.fh.band_name = 'B13'
        rad = self.fh.calibrate(data=self.counts, calibration='radiance').compute()
        rad_exp = np.array([[15.2, 12.],
                            [8.8, -0.8]])
        self.assertTrue(np.allclose(rad, rad_exp))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utilties for getting various angles for a dataset.."""
from __future__ import annotations

import hashlib
import os
import shutil
from datetime import datetime
from functools import update_wrapper
from glob import glob
from typing import Any, Callable, Optional, Union

import dask
import numpy as np
import xarray as xr
from dask import array as da
from pyorbital.astronomy import cos_zen as pyob_cos_zen
from pyorbital.astronomy import get_alt_az
from pyorbital.orbital import get_observer_look
from pyresample.geometry import AreaDefinition, StackedAreaDefinition, SwathDefinition

import satpy
from satpy.utils import get_satpos, ignore_invalid_float_warnings

PRGeometry = Union[SwathDefinition, AreaDefinition, StackedAreaDefinition]

# Arbitrary time used when computing sensor angles that is passed to
# pyorbital's get_observer_look function.
# The difference is on the order of 1e-10 at most as time changes so we force
# it to a single time for easier caching. It is *only* used if caching.
STATIC_EARTH_INERTIAL_DATETIME = datetime(2000, 1, 1, 12, 0, 0)
DEFAULT_UNCACHE_TYPES = (SwathDefinition, xr.DataArray, da.Array)
HASHABLE_GEOMETRIES = (AreaDefinition, StackedAreaDefinition)


class ZarrCacheHelper:
    """Helper for caching function results to on-disk zarr arrays.

    It is recommended to use this class through the :func:`cache_to_zarr_if`
    decorator rather than using it directly.

    Currently the cache does not perform any limiting or removal of cache
    content. That is left up to the user to manage. Caching is based on
    arguments passed to the decorated function but will only be performed
    if the arguments are of a certain type (see ``uncacheable_arg_types``).
    The cache value to use is purely based on the hash value of all of the
    provided arguments along with the "cache version" (see below).

    Args:
        func: Function that will be called to generate the value to cache.
        cache_config_key: Name of the boolean ``satpy.config`` parameter to
            use to determine if caching should be done.
        uncacheable_arg_types: Types that if present in the passed arguments
            should trigger caching to *not* happen. By default this includes
            ``SwathDefinition``, ``xr.DataArray``, and ``da.Array`` objects.
        sanitize_args_func: Optional function to call to sanitize provided
            arguments before they are considered for caching. This can be used
            to make arguments more "cacheable" by replacing them with similar
            values that will result in more cache hits. Note that the sanitized
            arguments are only passed to the underlying function if caching
            will be performed, otherwise the original arguments are passed.
        cache_version: Version number used to distinguish one version of a
            decorated function from future versions.

    Notes:
        * Caching only supports dask array values.

        * This helper allows for an additional ``cache_dir`` parameter to
          override the use of the ``satpy.config`` ``cache_dir`` parameter.

    Examples:
        To use through the :func:`cache_to_zarr_if` decorator::

            @cache_to_zarr_if("cache_my_stuff")
            def generate_my_stuff(area_def: AreaDefinition, some_factor: int) -> da.Array:
                # Generate
                return my_dask_arr

        To use the decorated function::

            with satpy.config.set(cache_my_stuff=True):
                my_stuff = generate_my_stuff(area_def, 5)

    """

    def __init__(self,
                 func: Callable,
                 cache_config_key: str,
                 uncacheable_arg_types=DEFAULT_UNCACHE_TYPES,
                 sanitize_args_func: Callable = None,
                 cache_version: int = 1,
                 ):
        """Hold on to provided arguments for future use."""
        self._func = func
        self._cache_config_key = cache_config_key
        self._uncacheable_arg_types = uncacheable_arg_types
        self._sanitize_args_func = sanitize_args_func
        self._cache_version = cache_version

    def cache_clear(self, cache_dir: Optional[str] = None):
        """Remove all on-disk files associated with this function.

        Intended to mimic the :func:`functools.cache` behavior.
        """
        if cache_dir is None:
            cache_dir = satpy.config.get("cache_dir")
        if cache_dir is None:
            raise RuntimeError("No 'cache_dir' configured.")
        zarr_pattern = self._zarr_pattern("*", cache_version="*").format("*")
        for zarr_dir in glob(os.path.join(cache_dir, zarr_pattern)):
            try:
                shutil.rmtree(zarr_dir)
            except OSError:
                continue

    def _zarr_pattern(self, arg_hash, cache_version: Union[int, str] = None) -> str:
        if cache_version is None:
            cache_version = self._cache_version
        return f"{self._func.__name__}_v{cache_version}" + "_{}_" + f"{arg_hash}.zarr"

    def __call__(self, *args, cache_dir: Optional[str] = None) -> Any:
        """Call the decorated function."""
        new_args = self._sanitize_args_func(*args) if self._sanitize_args_func is not None else args
        arg_hash = _hash_args(*new_args, unhashable_types=self._uncacheable_arg_types)
        should_cache, cache_dir = self._get_should_cache_and_cache_dir(new_args, cache_dir)
        zarr_fn = self._zarr_pattern(arg_hash)
        zarr_format = os.path.join(cache_dir, zarr_fn)
        zarr_paths = glob(zarr_format.format("*"))
        if not should_cache or not zarr_paths:
            # use sanitized arguments if we are caching, otherwise use original arguments
            args = new_args if should_cache else args
            res = self._func(*args)
            if should_cache and not zarr_paths:
                self._cache_results(res, zarr_format)
        # if we did any caching, let's load from the zarr files
        if should_cache:
            # re-calculate the cached paths
            zarr_paths = glob(zarr_format.format("*"))
            if not zarr_paths:
                raise RuntimeError("Data was cached to disk but no files were found")
            res = tuple(da.from_zarr(zarr_path) for zarr_path in zarr_paths)
        return res

    def _get_should_cache_and_cache_dir(self, args, cache_dir: Optional[str]) -> tuple[bool, str]:
        should_cache: bool = satpy.config.get(self._cache_config_key, False)
        can_cache = not any(isinstance(arg, self._uncacheable_arg_types) for arg in args)
        should_cache = should_cache and can_cache
        if cache_dir is None:
            cache_dir = satpy.config.get("cache_dir")
        return should_cache, cache_dir

    def _cache_results(self, res, zarr_format):
        os.makedirs(os.path.dirname(zarr_format), exist_ok=True)
        new_res = []
        for idx, sub_res in enumerate(res):
            if not isinstance(sub_res, da.Array):
                raise ValueError("Zarr caching currently only supports dask "
                                 f"arrays. Got {type(sub_res)}")
            zarr_path = zarr_format.format(idx)
            # See https://github.com/dask/dask/issues/8380
            with dask.config.set({"optimization.fuse.active": False}):
                new_sub_res = sub_res.to_zarr(zarr_path,
                                              return_stored=True,
                                              compute=False)
            new_res.append(new_sub_res)
        # actually compute the storage to zarr
        da.compute(new_res)


def cache_to_zarr_if(
        cache_config_key: str,
        uncacheable_arg_types=DEFAULT_UNCACHE_TYPES,
        sanitize_args_func: Callable = None,
) -> Callable:
    """Decorate a function and cache the results as a zarr array on disk.

    This only happens if the ``satpy.config`` boolean value for the provided
    key is ``True`` as well as some other conditions. See
    :class:`ZarrCacheHelper` for more information. Most importantly, this
    decorator does not limit how many items can be cached and does not clear
    out old entries. It is up to the user to manage the size of the cache.

    """
    def _decorator(func: Callable) -> Callable:
        zarr_cacher = ZarrCacheHelper(func,
                                      cache_config_key,
                                      uncacheable_arg_types,
                                      sanitize_args_func)
        wrapper = update_wrapper(zarr_cacher, func)
        return wrapper
    return _decorator


def _hash_args(*args, unhashable_types=DEFAULT_UNCACHE_TYPES):
    import json
    hashable_args = []
    for arg in args:
        if isinstance(arg, unhashable_types):
            continue
        if isinstance(arg, HASHABLE_GEOMETRIES):
            arg = hash(arg)
        elif isinstance(arg, datetime):
            arg = arg.isoformat(" ")
        hashable_args.append(arg)
    arg_hash = hashlib.sha1()  # nosec
    arg_hash.update(json.dumps(tuple(hashable_args)).encode('utf8'))
    return arg_hash.hexdigest()


def _sanitize_observer_look_args(*args):
    new_args = []
    for arg in args:
        if isinstance(arg, datetime):
            new_args.append(STATIC_EARTH_INERTIAL_DATETIME)
        elif isinstance(arg, (float, np.float64, np.float32)):
            # round floating point numbers to nearest tenth
            new_args.append(round(arg, 1))
        else:
            new_args.append(arg)
    return new_args


def _geo_dask_to_data_array(arr: da.Array) -> xr.DataArray:
    return xr.DataArray(arr, dims=('y', 'x'))


def get_angles(data_arr: xr.DataArray) -> tuple[xr.DataArray, xr.DataArray, xr.DataArray, xr.DataArray]:
    """Get sun and satellite azimuth and zenith angles.

    Note that this function can benefit from the ``satpy.config`` parameters
    :ref:`cache_lonlats <config_cache_lonlats_setting>` and
    :ref:`cache_sensor_angles <config_cache_sensor_angles_setting>`
    being set to ``True``.

    Args:
        data_arr: DataArray to get angles for. Information extracted from this
            object are ``.attrs["area"]`` and ``.attrs["start_time"]``.
            Additionally, the dask array chunk size is used when generating
            new arrays. The actual data of the object is not used.

    Returns:
        Four DataArrays representing sensor azimuth angle, sensor zenith angle,
        solar azimuth angle, and solar zenith angle. All values are in degrees.

    """
    sata, satz = _get_sensor_angles(data_arr)
    suna, sunz = _get_sun_angles(data_arr)
    return sata, satz, suna, sunz


def get_satellite_zenith_angle(data_arr: xr.DataArray) -> xr.DataArray:
    """Generate satellite zenith angle for the provided data.

    Note that this function can benefit from the ``satpy.config`` parameters
    :ref:`cache_lonlats <config_cache_lonlats_setting>` and
    :ref:`cache_sensor_angles <config_cache_sensor_angles_setting>`
    being set to ``True``. Values are in degrees.

    """
    satz = _get_sensor_angles(data_arr)[1]
    return satz


def get_cos_sza(data_arr: xr.DataArray) -> xr.DataArray:
    """Generate the cosine of the solar zenith angle for the provided data.

    Returns:
        DataArray with the same shape as ``data_arr``.

    """
    lons, lats = _get_valid_lonlats(data_arr.attrs["area"], data_arr.chunks)
    cos_sza = _get_cos_sza(data_arr.attrs["start_time"], lons, lats)
    return _geo_dask_to_data_array(cos_sza)


@cache_to_zarr_if("cache_lonlats")
def _get_valid_lonlats(area: PRGeometry, chunks: Union[int, str, tuple] = "auto") -> tuple[da.Array, da.Array]:
    with ignore_invalid_float_warnings():
        lons, lats = area.get_lonlats(chunks=chunks)
        lons = da.where(lons >= 1e30, np.nan, lons)
        lats = da.where(lats >= 1e30, np.nan, lats)
    return lons, lats


def _get_sun_angles(data_arr: xr.DataArray) -> tuple[xr.DataArray, xr.DataArray]:
    lons, lats = _get_valid_lonlats(data_arr.attrs["area"], data_arr.data.chunks)
    suna = da.map_blocks(_get_sun_azimuth_ndarray, lons, lats,
                         data_arr.attrs["start_time"],
                         dtype=lons.dtype, meta=np.array((), dtype=lons.dtype),
                         chunks=lons.chunks)
    cos_sza = _get_cos_sza(data_arr.attrs["start_time"], lons, lats)
    sunz = np.rad2deg(np.arccos(cos_sza))
    suna = _geo_dask_to_data_array(suna)
    sunz = _geo_dask_to_data_array(sunz)
    return suna, sunz


def _get_cos_sza(utc_time, lons, lats):
    cos_sza = da.map_blocks(_cos_zen_ndarray,
                            lons, lats, utc_time,
                            meta=np.array((), dtype=lons.dtype),
                            dtype=lons.dtype,
                            chunks=lons.chunks)
    return cos_sza


def _cos_zen_ndarray(lons, lats, utc_time):
    with ignore_invalid_float_warnings():
        return pyob_cos_zen(utc_time, lons, lats)


def _get_sun_azimuth_ndarray(lons: np.ndarray, lats: np.ndarray, start_time: datetime) -> np.ndarray:
    with ignore_invalid_float_warnings():
        suna = get_alt_az(start_time, lons, lats)[1]
        suna = np.rad2deg(suna)
    return suna


def _get_sensor_angles(data_arr: xr.DataArray) -> tuple[xr.DataArray, xr.DataArray]:
    sat_lon, sat_lat, sat_alt = get_satpos(data_arr)
    area_def = data_arr.attrs["area"]
    sata, satz = _get_sensor_angles_from_sat_pos(sat_lon, sat_lat, sat_alt,
                                                 data_arr.attrs["start_time"],
                                                 area_def, data_arr.data.chunks)
    sata = _geo_dask_to_data_array(sata)
    satz = _geo_dask_to_data_array(satz)
    return sata, satz


@cache_to_zarr_if("cache_sensor_angles", sanitize_args_func=_sanitize_observer_look_args)
def _get_sensor_angles_from_sat_pos(sat_lon, sat_lat, sat_alt, start_time, area_def, chunks):
    lons, lats = _get_valid_lonlats(area_def, chunks)
    res = da.map_blocks(_get_sensor_angles_ndarray, lons, lats, start_time, sat_lon, sat_lat, sat_alt,
                        dtype=lons.dtype, meta=np.array((), dtype=lons.dtype), new_axis=[0],
                        chunks=(2,) + lons.chunks)
    return res[0], res[1]


def _get_sensor_angles_ndarray(lons, lats, start_time, sat_lon, sat_lat, sat_alt) -> np.ndarray:
    with ignore_invalid_float_warnings():
        sata, satel = get_observer_look(
            sat_lon,
            sat_lat,
            sat_alt / 1000.0,  # km
            start_time,
            lons, lats, 0)
        satz = 90 - satel
        return np.stack([sata, satz])


def sunzen_corr_cos(data: da.Array,
                    cos_zen: da.Array,
                    limit: float = 88.,
                    max_sza: Optional[float] = 95.) -> da.Array:
    """Perform Sun zenith angle correction.

    The correction is based on the provided cosine of the zenith
    angle (``cos_zen``).  The correction is limited
    to ``limit`` degrees (default: 88.0 degrees).  For larger zenith
    angles, the correction is the same as at the ``limit`` if ``max_sza``
    is `None`. The default behavior is to gradually reduce the correction
    past ``limit`` degrees up to ``max_sza`` where the correction becomes
    0. Both ``data`` and ``cos_zen`` should be 2D arrays of the same shape.

    """
    return da.map_blocks(_sunzen_corr_cos_ndarray,
                         data, cos_zen, limit, max_sza,
                         meta=np.array((), dtype=data.dtype),
                         chunks=data.chunks)


def _sunzen_corr_cos_ndarray(data: np.ndarray,
                             cos_zen: np.ndarray,
                             limit: float,
                             max_sza: Optional[float]) -> np.ndarray:
    # Convert the zenith angle limit to cosine of zenith angle
    limit_rad = np.deg2rad(limit)
    limit_cos = np.cos(limit_rad)
    max_sza_rad = np.deg2rad(max_sza) if max_sza is not None else max_sza

    # Cosine correction
    corr = 1. / cos_zen
    if max_sza is not None:
        # gradually fall off for larger zenith angle
        grad_factor = (np.arccos(cos_zen) - limit_rad) / (max_sza_rad - limit_rad)
        # invert the factor so maximum correction is done at `limit` and falls off later
        grad_factor = 1. - np.log(grad_factor + 1) / np.log(2)
        # make sure we don't make anything negative
        grad_factor = grad_factor.clip(0.)
    else:
        # Use constant value (the limit) for larger zenith angles
        grad_factor = 1.
    corr = np.where(cos_zen > limit_cos, corr, grad_factor / limit_cos)
    # Force "night" pixels to 0 (where SZA is invalid)
    corr[np.isnan(cos_zen)] = 0
    return data * corr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit testing the enhancements functions, e.g. cira_stretch."""

import os
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import xarray as xr


class TestEnhancementStretch(unittest.TestCase):
    """Class for testing enhancements in satpy.enhancements."""

    def setUp(self):
        """Create test data used by every test."""
        data = np.arange(-210, 790, 100).reshape((2, 5)) * 0.95
        data[0, 0] = np.nan  # one bad value for testing
        crefl_data = np.arange(-210, 790, 100).reshape((2, 5)) * 0.95
        crefl_data /= 5.605
        crefl_data[0, 0] = np.nan  # one bad value for testing
        crefl_data[0, 1] = 0.
        self.ch1 = xr.DataArray(data, dims=('y', 'x'), attrs={'test': 'test'})
        self.ch2 = xr.DataArray(crefl_data, dims=('y', 'x'), attrs={'test': 'test'})
        rgb_data = np.stack([data, data, data])
        self.rgb = xr.DataArray(rgb_data, dims=('bands', 'y', 'x'),
                                coords={'bands': ['R', 'G', 'B']})

    def _test_enhancement(self, func, data, expected, **kwargs):
        """Perform basic checks that apply to multiple tests."""
        from trollimage.xrimage import XRImage

        pre_attrs = data.attrs
        img = XRImage(data)
        func(img, **kwargs)

        self.assertIsInstance(img.data.data, da.Array)
        self.assertListEqual(sorted(pre_attrs.keys()),
                             sorted(img.data.attrs.keys()),
                             "DataArray attributes were not preserved")

        np.testing.assert_allclose(img.data.values, expected, atol=1.e-6, rtol=0)

    def test_cira_stretch(self):
        """Test applying the cira_stretch."""
        from satpy.enhancements import cira_stretch

        expected = np.array([[
            [np.nan, -7.04045974, -7.04045974, 0.79630132, 0.95947296],
            [1.05181359, 1.11651012, 1.16635571, 1.20691137, 1.24110186]]])
        self._test_enhancement(cira_stretch, self.ch1, expected)

    def test_reinhard(self):
        """Test the reinhard algorithm."""
        from satpy.enhancements import reinhard_to_srgb
        expected = np.array([[[np.nan, 0., 0., 0.93333793, 1.29432402],
                              [1.55428709, 1.76572249, 1.94738635, 2.10848544, 2.25432809]],

                             [[np.nan, 0., 0., 0.93333793, 1.29432402],
                              [1.55428709, 1.76572249, 1.94738635, 2.10848544, 2.25432809]],

                             [[np.nan, 0., 0., 0.93333793, 1.29432402],
                              [1.55428709, 1.76572249, 1.94738635, 2.10848544, 2.25432809]]])
        self._test_enhancement(reinhard_to_srgb, self.rgb, expected)

    def test_lookup(self):
        """Test the lookup enhancement function."""
        from satpy.enhancements import lookup
        expected = np.array([[
            [0., 0., 0., 0.333333, 0.705882],
            [1., 1., 1., 1., 1.]]])
        lut = np.arange(256.)
        self._test_enhancement(lookup, self.ch1, expected, luts=lut)

        expected = np.array([[[0., 0., 0., 0.333333, 0.705882],
                              [1., 1., 1., 1., 1.]],
                             [[0., 0., 0., 0.333333, 0.705882],
                              [1., 1., 1., 1., 1.]],
                             [[0., 0., 0., 0.333333, 0.705882],
                              [1., 1., 1., 1., 1.]]])
        lut = np.arange(256.)
        lut = np.vstack((lut, lut, lut)).T
        self._test_enhancement(lookup, self.rgb, expected, luts=lut)

    def test_colorize(self):
        """Test the colorize enhancement function."""
        from trollimage.colormap import brbg

        from satpy.enhancements import colorize
        expected = np.array([[
            [np.nan, 3.29409498e-01, 3.29409498e-01,
             4.35952940e-06, 4.35952940e-06],
            [4.35952940e-06, 4.35952940e-06, 4.35952940e-06,
             4.35952940e-06, 4.35952940e-06]],
            [[np.nan, 1.88249866e-01, 1.88249866e-01,
              2.35302110e-01, 2.35302110e-01],
             [2.35302110e-01, 2.35302110e-01, 2.35302110e-01,
              2.35302110e-01, 2.35302110e-01]],
            [[np.nan, 1.96102817e-02, 1.96102817e-02,
              1.88238767e-01, 1.88238767e-01],
             [1.88238767e-01, 1.88238767e-01, 1.88238767e-01,
              1.88238767e-01, 1.88238767e-01]]])
        self._test_enhancement(colorize, self.ch1, expected, palettes=brbg)

    def test_palettize(self):
        """Test the palettize enhancement function."""
        from trollimage.colormap import brbg

        from satpy.enhancements import palettize
        expected = np.array([[[10, 0, 0, 10, 10], [10, 10, 10, 10, 10]]])
        self._test_enhancement(palettize, self.ch1, expected, palettes=brbg)

    def test_three_d_effect(self):
        """Test the three_d_effect enhancement function."""
        from satpy.enhancements import three_d_effect
        expected = np.array([[
            [np.nan, np.nan, -389.5, -294.5, 826.5],
            [np.nan, np.nan, 85.5, 180.5, 1301.5]]])
        self._test_enhancement(three_d_effect, self.ch1, expected)

    def test_crefl_scaling(self):
        """Test the crefl_scaling enhancement function."""
        from satpy.enhancements import crefl_scaling
        expected = np.array([[
            [np.nan, 0., 0., 0.44378, 0.631734],
            [0.737562, 0.825041, 0.912521, 1., 1.]]])
        self._test_enhancement(crefl_scaling, self.ch2, expected, idx=[0., 25., 55., 100., 255.],
                               sc=[0., 90., 140., 175., 255.])

    def test_piecewise_linear_stretch(self):
        """Test the piecewise_linear_stretch enhancement function."""
        from satpy.enhancements import piecewise_linear_stretch
        expected = np.array([[
            [np.nan, 0., 0., 0.44378, 0.631734],
            [0.737562, 0.825041, 0.912521, 1., 1.]]])
        self._test_enhancement(piecewise_linear_stretch,
                               self.ch2 / 100.0,
                               expected,
                               xp=[0., 25., 55., 100., 255.],
                               fp=[0., 90., 140., 175., 255.],
                               reference_scale_factor=255,
                               )

    def test_btemp_threshold(self):
        """Test applying the cira_stretch."""
        from satpy.enhancements import btemp_threshold

        expected = np.array([[
            [np.nan, 0.946207, 0.892695, 0.839184, 0.785672],
            [0.73216, 0.595869, 0.158745, -0.278379, -0.715503]]])
        self._test_enhancement(btemp_threshold, self.ch1, expected,
                               min_in=-200, max_in=500, threshold=350)

    def test_merge_colormaps(self):
        """Test merging colormaps."""
        from trollimage.colormap import Colormap

        from satpy.enhancements import _merge_colormaps as mcp
        from satpy.enhancements import create_colormap
        ret_map = mock.MagicMock()

        create_colormap_mock = mock.Mock(wraps=create_colormap)
        cmap1 = Colormap((1, (1., 1., 1.)))
        kwargs = {'palettes': cmap1}

        with mock.patch('satpy.enhancements.create_colormap', create_colormap_mock):
            res = mcp(kwargs)
        self.assertTrue(res is cmap1)
        create_colormap_mock.assert_not_called()
        create_colormap_mock.reset_mock()
        ret_map.reset_mock()

        cmap1 = {'colors': 'blues', 'min_value': 0,
                 'max_value': 1}
        kwargs = {'palettes': [cmap1]}
        with mock.patch('satpy.enhancements.create_colormap', create_colormap_mock),\
                mock.patch('trollimage.colormap.blues', ret_map):
            _ = mcp(kwargs)
        create_colormap_mock.assert_called_once()
        ret_map.reverse.assert_not_called()
        ret_map.set_range.assert_called_with(0, 1)
        create_colormap_mock.reset_mock()
        ret_map.reset_mock()

        cmap2 = {'colors': 'blues', 'min_value': 2,
                 'max_value': 3, 'reverse': True}
        kwargs = {'palettes': [cmap2]}
        with mock.patch('trollimage.colormap.blues', ret_map):
            _ = mcp(kwargs)
        ret_map.reverse.assert_called_once()
        ret_map.set_range.assert_called_with(2, 3)
        create_colormap_mock.reset_mock()
        ret_map.reset_mock()

        kwargs = {'palettes': [cmap1, cmap2]}
        with mock.patch('trollimage.colormap.blues', ret_map):
            _ = mcp(kwargs)
        ret_map.__add__.assert_called_once()

    def tearDown(self):
        """Clean up."""


class TestColormapLoading(unittest.TestCase):
    """Test utilities used with colormaps."""

    def test_cmap_from_file_rgb(self):
        """Test that colormaps can be loaded from a binary file."""
        from tempfile import NamedTemporaryFile

        from satpy.enhancements import create_colormap

        # create the colormap file on disk
        with NamedTemporaryFile(suffix='.npy', delete=False) as tmp_cmap:
            cmap_filename = tmp_cmap.name
            np.save(cmap_filename, np.array([
                [255, 0, 0],
                [255, 255, 0],
                [255, 255, 255],
                [0, 0, 255],
            ]))

        try:
            cmap = create_colormap({'filename': cmap_filename})
            self.assertEqual(cmap.colors.shape[0], 4)
            np.testing.assert_equal(cmap.colors[0], [1.0, 0, 0])
            self.assertEqual(cmap.values.shape[0], 4)
            self.assertEqual(cmap.values[0], 0)
            self.assertEqual(cmap.values[-1], 1.0)

            cmap = create_colormap({'filename': cmap_filename, 'min_value': 50, 'max_value': 100})
            self.assertEqual(cmap.colors.shape[0], 4)
            np.testing.assert_equal(cmap.colors[0], [1.0, 0, 0])
            self.assertEqual(cmap.values.shape[0], 4)
            self.assertEqual(cmap.values[0], 50)
            self.assertEqual(cmap.values[-1], 100)
        finally:
            os.remove(cmap_filename)

    def test_cmap_from_file_rgb_1(self):
        """Test that colormaps can be loaded from a binary file with 0-1 colors."""
        from tempfile import NamedTemporaryFile

        from satpy.enhancements import create_colormap

        # create the colormap file on disk
        with NamedTemporaryFile(suffix='.npy', delete=False) as tmp_cmap:
            cmap_filename = tmp_cmap.name
            np.save(cmap_filename, np.array([
                [1, 0, 0],
                [1, 1, 0],
                [1, 1, 1],
                [0, 0, 1],
            ]))

        try:
            cmap = create_colormap({'filename': cmap_filename,
                                    'color_scale': 1})
            self.assertEqual(cmap.colors.shape[0], 4)
            np.testing.assert_equal(cmap.colors[0], [1.0, 0, 0])
            self.assertEqual(cmap.values.shape[0], 4)
            self.assertEqual(cmap.values[0], 0)
            self.assertEqual(cmap.values[-1], 1.0)

            cmap = create_colormap({'filename': cmap_filename, 'color_scale': 1,
                                    'min_value': 50, 'max_value': 100})
            self.assertEqual(cmap.colors.shape[0], 4)
            np.testing.assert_equal(cmap.colors[0], [1.0, 0, 0])
            self.assertEqual(cmap.values.shape[0], 4)
            self.assertEqual(cmap.values[0], 50)
            self.assertEqual(cmap.values[-1], 100)
        finally:
            os.remove(cmap_filename)

    def test_cmap_from_file_vrgb(self):
        """Test that colormaps can be loaded from a binary file with values."""
        from tempfile import NamedTemporaryFile

        from satpy.enhancements import create_colormap

        # create the colormap file on disk
        with NamedTemporaryFile(suffix='.npy', delete=False) as tmp_cmap:
            cmap_filename = tmp_cmap.name
            np.save(cmap_filename, np.array([
                [128, 255, 0, 0],
                [130, 255, 255, 0],
                [132, 255, 255, 255],
                [134, 0, 0, 255],
            ]))

        try:
            # default mode of VRGB
            cmap = create_colormap({'filename': cmap_filename})
            self.assertEqual(cmap.colors.shape[0], 4)
            np.testing.assert_equal(cmap.colors[0], [1.0, 0, 0])
            self.assertEqual(cmap.values.shape[0], 4)
            self.assertEqual(cmap.values[0], 128)
            self.assertEqual(cmap.values[-1], 134)

            cmap = create_colormap({'filename': cmap_filename, 'colormap_mode': 'RGBA'})
            self.assertEqual(cmap.colors.shape[0], 4)
            self.assertEqual(cmap.colors.shape[1], 4)  # RGBA
            np.testing.assert_equal(cmap.colors[0], [128 / 255., 1.0, 0, 0])
            self.assertEqual(cmap.values.shape[0], 4)
            self.assertEqual(cmap.values[0], 0)
            self.assertEqual(cmap.values[-1], 1.0)

            cmap = create_colormap({'filename': cmap_filename, 'min_value': 50, 'max_value': 100})
            self.assertEqual(cmap.colors.shape[0], 4)
            np.testing.assert_equal(cmap.colors[0], [1.0, 0, 0])
            self.assertEqual(cmap.values.shape[0], 4)
            self.assertEqual(cmap.values[0], 50)
            self.assertEqual(cmap.values[-1], 100)

            self.assertRaises(ValueError, create_colormap,
                              {'filename': cmap_filename, 'colormap_mode': 'RGB',
                               'min_value': 50, 'max_value': 100})
        finally:
            os.remove(cmap_filename)

    def test_cmap_from_file_vrgba(self):
        """Test that colormaps can be loaded RGBA colors and values."""
        from tempfile import NamedTemporaryFile

        from satpy.enhancements import create_colormap

        # create the colormap file on disk
        with NamedTemporaryFile(suffix='.npy', delete=False) as tmp_cmap:
            cmap_filename = tmp_cmap.name
            np.save(cmap_filename, np.array([
                [128, 128, 255, 0, 0],  # value, R, G, B, A
                [130, 130, 255, 255, 0],
                [132, 132, 255, 255, 255],
                [134, 134, 0, 0, 255],
            ]))

        try:
            # default mode of VRGBA
            cmap = create_colormap({'filename': cmap_filename})
            self.assertEqual(cmap.colors.shape[0], 4)
            self.assertEqual(cmap.colors.shape[1], 4)  # RGBA
            np.testing.assert_equal(cmap.colors[0], [128 / 255.0, 1.0, 0, 0])
            self.assertEqual(cmap.values.shape[0], 4)
            self.assertEqual(cmap.values[0], 128)
            self.assertEqual(cmap.values[-1], 134)

            self.assertRaises(ValueError, create_colormap,
                              {'filename': cmap_filename, 'colormap_mode': 'RGBA'})

            cmap = create_colormap({'filename': cmap_filename, 'min_value': 50, 'max_value': 100})
            self.assertEqual(cmap.colors.shape[0], 4)
            self.assertEqual(cmap.colors.shape[1], 4)  # RGBA
            np.testing.assert_equal(cmap.colors[0], [128 / 255.0, 1.0, 0, 0])
            self.assertEqual(cmap.values.shape[0], 4)
            self.assertEqual(cmap.values[0], 50)
            self.assertEqual(cmap.values[-1], 100)
        finally:
            os.remove(cmap_filename)

    def test_cmap_from_file_bad_shape(self):
        """Test that unknown array shape causes an error."""
        from tempfile import NamedTemporaryFile

        from satpy.enhancements import create_colormap

        # create the colormap file on disk
        with NamedTemporaryFile(suffix='.npy', delete=False) as tmp_cmap:
            cmap_filename = tmp_cmap.name
            np.save(cmap_filename, np.array([
                [0],
                [64],
                [128],
                [255],
            ]))

        try:
            self.assertRaises(ValueError, create_colormap,
                              {'filename': cmap_filename})
        finally:
            os.remove(cmap_filename)

    def test_cmap_from_trollimage(self):
        """Test that colormaps in trollimage can be loaded."""
        from satpy.enhancements import create_colormap
        cmap = create_colormap({'colors': 'pubu'})
        from trollimage.colormap import pubu
        np.testing.assert_equal(cmap.colors, pubu.colors)
        np.testing.assert_equal(cmap.values, pubu.values)

    def test_cmap_no_colormap(self):
        """Test that being unable to create a colormap raises an error."""
        from satpy.enhancements import create_colormap
        self.assertRaises(ValueError, create_colormap, {})

    def test_cmap_list(self):
        """Test that colors can be a list/tuple."""
        from satpy.enhancements import create_colormap
        colors = [
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ]
        values = [2, 4, 6, 8]
        cmap = create_colormap({'colors': colors, 'color_scale': 1})
        self.assertEqual(cmap.colors.shape[0], 4)
        np.testing.assert_equal(cmap.colors[0], [0.0, 0.0, 1.0])
        self.assertEqual(cmap.values.shape[0], 4)
        self.assertEqual(cmap.values[0], 0)
        self.assertEqual(cmap.values[-1], 1.0)

        cmap = create_colormap({'colors': colors, 'color_scale': 1, 'values': values})
        self.assertEqual(cmap.colors.shape[0], 4)
        np.testing.assert_equal(cmap.colors[0], [0.0, 0.0, 1.0])
        self.assertEqual(cmap.values.shape[0], 4)
        self.assertEqual(cmap.values[0], 2)
        self.assertEqual(cmap.values[-1], 8)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The hrit ahi reader tests package."""

import unittest
from unittest import mock

import dask.array as da
import numpy as np
from xarray import DataArray


class TestHRITJMAFileHandler(unittest.TestCase):
    """Test the HRITJMAFileHandler."""

    @mock.patch('satpy.readers.hrit_jma.HRITFileHandler.__init__')
    def _get_reader(self, mocked_init, mda, filename_info=None):
        from satpy.readers.hrit_jma import HRITJMAFileHandler
        if not filename_info:
            filename_info = {}
        HRITJMAFileHandler.filename = 'filename'
        HRITJMAFileHandler.mda = mda
        return HRITJMAFileHandler('filename', filename_info, {})

    def _get_acq_time(self, nlines):
        """Get sample header entry for scanline acquisition times.

        Lines: 1, 21, 41, 61, ..., nlines
        Times: 1970-01-01 00:00 + (1, 21, 41, 61, ..., nlines) seconds

        So the interpolated times are expected to be 1970-01-01 +
        (1, 2, 3, 4, ..., nlines) seconds. Note that there will be some
        floating point inaccuracies, because timestamps are stored
        with only 6 decimals precision.
        """
        mjd_1970 = 40587.0
        lines_sparse = np.array(list(range(1, nlines, 20)) + [nlines])
        times_sparse = mjd_1970 + lines_sparse / 24 / 3600
        acq_time_s = ['LINE:={}\rTIME:={:.6f}\r'.format(line, time)
                      for line, time in zip(lines_sparse, times_sparse)]
        acq_time_b = ''.join(acq_time_s).encode()
        return acq_time_b

    def _get_mda(self, loff=5500.0, coff=5500.0, nlines=11000, ncols=11000,
                 segno=0, numseg=1, vis=True, platform='Himawari-8'):
        """Create metadata dict like HRITFileHandler would do it."""
        if vis:
            idf = b'$HALFTONE:=16\r_NAME:=VISIBLE\r_UNIT:=ALBEDO(%)\r' \
                  b'0:=-0.10\r1023:=100.00\r65535:=100.00\r'
        else:
            idf = b'$HALFTONE:=16\r_NAME:=INFRARED\r_UNIT:=KELVIN\r' \
                  b'0:=329.98\r1023:=130.02\r65535:=130.02\r'
        proj_h8 = b'GEOS(140.70)                    '
        proj_mtsat2 = b'GEOS(145.00)                    '
        proj_name = proj_h8 if platform == 'Himawari-8' else proj_mtsat2
        return {'image_segm_seq_no': segno,
                'total_no_image_segm': numseg,
                'projection_name': proj_name,
                'projection_parameters': {
                    'a': 6378169.00,
                    'b': 6356583.80,
                    'h': 35785831.00,
                },
                'cfac': 10233128,
                'lfac': 10233128,
                'coff': coff,
                'loff': loff,
                'number_of_columns': ncols,
                'number_of_lines': nlines,
                'image_data_function': idf,
                'image_observation_time': self._get_acq_time(nlines)}

    def test_init(self):
        """Test creating the file handler."""
        from satpy.readers.hrit_jma import HIMAWARI8, UNKNOWN_AREA

        # Test addition of extra metadata
        mda = self._get_mda()
        mda_expected = mda.copy()
        mda_expected.update(
            {'planned_end_segment_number': 1,
             'planned_start_segment_number': 1,
             'segment_sequence_number': 0,
             'unit': 'ALBEDO(%)'})
        mda_expected['projection_parameters']['SSP_longitude'] = 140.7
        reader = self._get_reader(mda=mda)
        self.assertEqual(reader.mda, mda_expected)

        # Check projection name
        self.assertEqual(reader.projection_name, 'GEOS(140.70)')

        # Check calibration table
        cal_expected = np.array([[0, -0.1],
                                 [1023,  100],
                                 [65535,  100]])
        self.assertTrue(np.all(reader.calibration_table == cal_expected))

        # Check if scanline timestamps are there (dedicated test below)
        self.assertIsInstance(reader.acq_time, np.ndarray)

        # Check platform
        self.assertEqual(reader.platform, HIMAWARI8)

        # Check is_segmented attribute
        expected = {0: False, 1: True, 8: True}
        for segno, is_segmented in expected.items():
            mda = self._get_mda(segno=segno)
            reader = self._get_reader(mda=mda)
            self.assertEqual(reader.is_segmented, is_segmented)

        # Check area IDs
        expected = [
            ({'area': 1}, 1),
            ({'area': 1234}, UNKNOWN_AREA),
            ({}, UNKNOWN_AREA)
        ]
        mda = self._get_mda()
        for filename_info, area_id in expected:
            reader = self._get_reader(mda=mda, filename_info=filename_info)
            self.assertEqual(reader.area_id, area_id)

    @mock.patch('satpy.readers.hrit_jma.HRITJMAFileHandler.__init__')
    def test_get_platform(self, mocked_init):
        """Test platform identification."""
        from satpy.readers.hrit_jma import PLATFORMS, UNKNOWN_PLATFORM, HRITJMAFileHandler

        mocked_init.return_value = None
        reader = HRITJMAFileHandler()

        for proj_name, platform in PLATFORMS.items():
            reader.projection_name = proj_name
            self.assertEqual(reader._get_platform(), platform)

        with mock.patch('logging.Logger.error') as mocked_log:
            reader.projection_name = 'invalid'
            self.assertEqual(reader._get_platform(), UNKNOWN_PLATFORM)
            mocked_log.assert_called()

    def test_get_area_def(self):
        """Test getting an AreaDefinition."""
        from satpy.readers.hrit_jma import AREA_NAMES, FULL_DISK, NORTH_HEMIS, SOUTH_HEMIS

        cases = [
            # Non-segmented, full disk
            {'loff': 1375.0, 'coff': 1375.0,
             'nlines': 2750, 'ncols': 2750,
             'segno': 0, 'numseg': 1,
             'area': FULL_DISK,
             'extent': (-5498000.088960204, -5498000.088960204,
                        5502000.089024927, 5502000.089024927)},
            # Non-segmented, northern hemisphere
            {'loff': 1325.0, 'coff': 1375.0,
             'nlines': 1375, 'ncols': 2750,
             'segno': 0, 'numseg': 1,
             'area': NORTH_HEMIS,
             'extent': (-5498000.088960204, -198000.00320373234,
                        5502000.089024927, 5302000.085788833)},
            # Non-segmented, southern hemisphere
            {'loff': 50, 'coff': 1375.0,
             'nlines': 1375, 'ncols': 2750,
             'segno': 0, 'numseg': 1,
             'area': SOUTH_HEMIS,
             'extent': (-5498000.088960204, -5298000.085724112,
                        5502000.089024927, 202000.0032684542)},
            # Segmented, segment #1
            {'loff': 1375.0, 'coff': 1375.0,
             'nlines': 275, 'ncols': 2750,
             'segno': 1, 'numseg': 10,
             'area': FULL_DISK,
             'extent': (-5498000.088960204, 4402000.071226413,
                        5502000.089024927, 5502000.089024927)},
            # Segmented, segment #7
            {'loff': 1375.0, 'coff': 1375.0,
             'nlines': 275, 'ncols': 2750,
             'segno': 7, 'numseg': 10,
             'area': FULL_DISK,
             'extent': (-5498000.088960204, -2198000.035564665,
                        5502000.089024927, -1098000.0177661523)},
        ]
        for case in cases:
            mda = self._get_mda(loff=case['loff'], coff=case['coff'],
                                nlines=case['nlines'], ncols=case['ncols'],
                                segno=case['segno'], numseg=case['numseg'])
            reader = self._get_reader(mda=mda,
                                      filename_info={'area': case['area']})
            area = reader.get_area_def('some_id')
            self.assertTupleEqual(area.area_extent, case['extent'])
            self.assertEqual(area.description, AREA_NAMES[case['area']]['long'])

    def test_calibrate(self):
        """Test calibration."""
        # Generate test data
        counts = np.linspace(0, 1200, 25).reshape(5, 5)
        counts[-1, -1] = 65535
        counts = DataArray(da.from_array(counts, chunks=5))
        refl = np.array(
            [[-0.1,            4.79247312,   9.68494624,  14.57741935,  19.46989247],
             [24.36236559,  29.25483871,  34.14731183,  39.03978495,  43.93225806],
             [48.82473118,  53.7172043,   58.60967742,  63.50215054,  68.39462366],
             [73.28709677,  78.17956989,  83.07204301,  87.96451613,  92.85698925],
             [97.74946237,  100.,         100.,         100.,         np.nan]]
        )
        bt = np.array(
            [[329.98,            320.20678397, 310.43356794, 300.66035191, 290.88713587],
             [281.11391984, 271.34070381, 261.56748778, 251.79427175, 242.02105572],
             [232.24783969, 222.47462366, 212.70140762, 202.92819159, 193.15497556],
             [183.38175953, 173.6085435,  163.83532747, 154.06211144, 144.28889541],
             [134.51567937, 130.02,       130.02,       130.02,       np.nan]]
        )

        # Choose an area near the subsatellite point to avoid masking
        # of space pixels
        mda = self._get_mda(nlines=5, ncols=5, loff=1375.0, coff=1375.0,
                            segno=0)
        reader = self._get_reader(mda=mda)

        # 1. Counts
        res = reader.calibrate(data=counts, calibration='counts')
        self.assertTrue(np.all(counts.values == res.values))

        # 2. Reflectance
        res = reader.calibrate(data=counts, calibration='reflectance')
        np.testing.assert_allclose(refl, res.values)  # also compares NaN

        # 3. Brightness temperature
        mda_bt = self._get_mda(nlines=5, ncols=5, loff=1375.0, coff=1375.0,
                               segno=0, vis=False)
        reader_bt = self._get_reader(mda=mda_bt)
        res = reader_bt.calibrate(data=counts,
                                  calibration='brightness_temperature')
        np.testing.assert_allclose(bt, res.values)  # also compares NaN

    def test_mask_space(self):
        """Test masking of space pixels."""
        mda = self._get_mda(loff=1375.0, coff=1375.0, nlines=275, ncols=1375,
                            segno=1, numseg=10)
        reader = self._get_reader(mda=mda)
        data = DataArray(da.ones((275, 1375), chunks=1024))
        masked = reader._mask_space(data)

        # First line of the segment should be space, in the middle of the
        # last line there should be some valid pixels
        np.testing.assert_allclose(masked.values[0, :], np.nan)
        self.assertTrue(np.all(masked.values[-1, 588:788] == 1))

    @mock.patch('satpy.readers.hrit_jma.HRITFileHandler.get_dataset')
    def test_get_dataset(self, base_get_dataset):
        """Test getting a dataset."""
        from satpy.readers.hrit_jma import HIMAWARI8

        mda = self._get_mda(loff=1375.0, coff=1375.0, nlines=275, ncols=1375,
                            segno=1, numseg=10)
        reader = self._get_reader(mda=mda)

        key = mock.MagicMock()
        key.calibration = 'reflectance'

        base_get_dataset.return_value = DataArray(da.ones((275, 1375),
                                                          chunks=1024),
                                                  dims=('y', 'x'))

        # Check attributes
        res = reader.get_dataset(key, {'units': '%', 'sensor': 'ahi'})
        self.assertEqual(res.attrs['units'], '%')
        self.assertEqual(res.attrs['sensor'], 'ahi')
        self.assertEqual(res.attrs['platform_name'], HIMAWARI8)
        self.assertEqual(res.attrs['satellite_longitude'], 140.7)
        self.assertEqual(res.attrs['satellite_latitude'], 0.)
        self.assertEqual(res.attrs['satellite_altitude'], 35785831.0)
        self.assertDictEqual(res.attrs['orbital_parameters'], {'projection_longitude': 140.7,
                                                               'projection_latitude': 0.,
                                                               'projection_altitude': 35785831.0})

        # Check if acquisition time is a coordinate
        self.assertIn('acq_time', res.coords)

        # Check called methods
        with mock.patch.object(reader, '_mask_space') as mask_space:
            with mock.patch.object(reader, 'calibrate') as calibrate:
                reader.get_dataset(key, {'units': '%', 'sensor': 'ahi'})
                mask_space.assert_called()
                calibrate.assert_called()

        with mock.patch('logging.Logger.error') as log_mock:
            reader.get_dataset(key, {'units': '%', 'sensor': 'jami'})
            log_mock.assert_called()

    def test_mjd2datetime64(self):
        """Test conversion from modified julian day to datetime64."""
        from satpy.readers.hrit_jma import mjd2datetime64
        self.assertEqual(mjd2datetime64(np.array([0])),
                         np.datetime64('1858-11-17', 'us'))
        self.assertEqual(mjd2datetime64(np.array([40587.5])),
                         np.datetime64('1970-01-01 12:00', 'us'))

    def test_get_acq_time(self):
        """Test computation of scanline acquisition times."""
        dt_line = np.arange(1, 11000+1).astype('timedelta64[s]')
        acq_time_exp = np.datetime64('1970-01-01', 'us') + dt_line

        for platform in ['Himawari-8', 'MTSAT-2']:
            # Results are not exactly identical because timestamps are stored in
            # the header with only 6 decimals precision (max diff here: 45 msec).
            mda = self._get_mda(platform=platform)
            reader = self._get_reader(mda=mda)
            np.testing.assert_allclose(reader.acq_time.astype(np.int64),
                                       acq_time_exp.astype(np.int64),
                                       atol=45000)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Satpy Configuration directory and file handling."""

import ast
import glob
import logging
import os
import sys
from collections import OrderedDict

import appdirs
import pkg_resources
from donfig import Config

LOG = logging.getLogger(__name__)

BASE_PATH = os.path.dirname(os.path.realpath(__file__))
# FIXME: Use package_resources?
PACKAGE_CONFIG_PATH = os.path.join(BASE_PATH, 'etc')

_satpy_dirs = appdirs.AppDirs(appname='satpy', appauthor='pytroll')
_CONFIG_DEFAULTS = {
    'cache_dir': _satpy_dirs.user_cache_dir,
    'cache_lonlats': False,
    'cache_sensor_angles': False,
    'config_path': [],
    'data_dir': _satpy_dirs.user_data_dir,
    'demo_data_dir': '.',
    'download_aux': True,
}

# Satpy main configuration object
# See https://donfig.readthedocs.io/en/latest/configuration.html
# for more information.
#
# Configuration values will be loaded from files at:
# 1. The builtin package satpy.yaml (not present currently)
# 2. $SATPY_ROOT_CONFIG (default: /etc/satpy/satpy.yaml)
# 3. <python-env-prefix>/etc/satpy/satpy.yaml
# 4. ~/.config/satpy/satpy.yaml
# 5. ~/.satpy/satpy.yaml
# 6. $SATPY_CONFIG_PATH/satpy.yaml if present (colon separated)
_CONFIG_PATHS = [
    os.path.join(PACKAGE_CONFIG_PATH, 'satpy.yaml'),
    os.getenv('SATPY_ROOT_CONFIG', os.path.join('/etc', 'satpy', 'satpy.yaml')),
    os.path.join(sys.prefix, 'etc', 'satpy', 'satpy.yaml'),
    os.path.join(_satpy_dirs.user_config_dir, 'satpy.yaml'),
    os.path.join(os.path.expanduser('~'), '.satpy', 'satpy.yaml'),
]
# The above files can also be directories. If directories all files
# with `.yaml`., `.yml`, or `.json` extensions will be used.

_ppp_config_dir = os.getenv('PPP_CONFIG_DIR', None)
_satpy_config_path = os.getenv('SATPY_CONFIG_PATH', None)

if _ppp_config_dir is not None and _satpy_config_path is None:
    LOG.warning("'PPP_CONFIG_DIR' is deprecated. Please use 'SATPY_CONFIG_PATH' instead.")
    _satpy_config_path = _ppp_config_dir

if _satpy_config_path is not None:
    if _satpy_config_path.startswith("["):
        # 'SATPY_CONFIG_PATH' is set by previous satpy config as a reprsentation of a 'list'
        # need to use 'ast.literal_eval' to parse the string back to a list
        _satpy_config_path_list = ast.literal_eval(_satpy_config_path)
    else:
        # colon-separated are ordered by custom -> builtins
        # i.e. last-applied/highest priority to first-applied/lowest priority
        _satpy_config_path_list = _satpy_config_path.split(os.pathsep)

    os.environ['SATPY_CONFIG_PATH'] = repr(_satpy_config_path_list)
    for config_dir in _satpy_config_path_list:
        _CONFIG_PATHS.append(os.path.join(config_dir, 'satpy.yaml'))

_ancpath = os.getenv('SATPY_ANCPATH', None)
_data_dir = os.getenv('SATPY_DATA_DIR', None)
if _ancpath is not None and _data_dir is None:
    LOG.warning("'SATPY_ANCPATH' is deprecated. Please use 'SATPY_DATA_DIR' instead.")
    os.environ['SATPY_DATA_DIR'] = _ancpath

config = Config("satpy", defaults=[_CONFIG_DEFAULTS], paths=_CONFIG_PATHS)


def get_config_path_safe():
    """Get 'config_path' and check for proper 'list' type."""
    config_path = config.get('config_path')
    if not isinstance(config_path, list):
        raise ValueError("Satpy config option 'config_path' must be a "
                         "list, not '{}'".format(type(config_path)))
    return config_path


def get_entry_points_config_dirs(name, include_config_path=True):
    """Get the config directories for all entry points of given name."""
    dirs = []
    for entry_point in pkg_resources.iter_entry_points(name):
        package_name = entry_point.module_name.split('.', 1)[0]
        new_dir = os.path.join(entry_point.dist.module_path, package_name, 'etc')
        if not dirs or dirs[-1] != new_dir:
            dirs.append(new_dir)
    if include_config_path:
        dirs.extend(config.get('config_path')[::-1])
    return dirs


def config_search_paths(filename, search_dirs=None, **kwargs):
    """Get series of configuration base paths where Satpy configs are located."""
    if search_dirs is None:
        search_dirs = get_config_path_safe()[::-1]

    paths = [filename, os.path.basename(filename)]
    paths += [os.path.join(search_dir, filename) for search_dir in search_dirs]
    paths += [os.path.join(PACKAGE_CONFIG_PATH, filename)]
    paths = [os.path.abspath(path) for path in paths]

    if kwargs.get("check_exists", True):
        paths = [x for x in paths if os.path.isfile(x)]

    paths = list(OrderedDict.fromkeys(paths))
    # flip the order of the list so builtins are loaded first
    return paths[::-1]


def glob_config(pattern, search_dirs=None):
    """Return glob results for all possible configuration locations.

    Note: This method does not check the configuration "base" directory if the pattern includes a subdirectory.
          This is done for performance since this is usually used to find *all* configs for a certain component.
    """
    patterns = config_search_paths(pattern, search_dirs=search_dirs,
                                   check_exists=False)
    for pattern_fn in patterns:
        for path in glob.iglob(pattern_fn):
            yield path


def get_config_path(filename):
    """Get the path to the highest priority version of a config file."""
    paths = config_search_paths(filename)
    for path in paths[::-1]:
        if os.path.exists(path):
            return path
    raise FileNotFoundError("Could not find file in configuration path: "
                            "'{}'".format(filename))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The HRIT electrol reader tests package."""

import datetime
import unittest
from unittest import mock

import dask.array as da
import numpy as np
from xarray import DataArray

from satpy.readers.electrol_hrit import (
    HRITGOMSEpilogueFileHandler,
    HRITGOMSFileHandler,
    HRITGOMSPrologueFileHandler,
    epilogue,
    image_acquisition,
    prologue,
    recarray2dict,
    satellite_status,
)
from satpy.tests.utils import make_dataid

# Simplify some type selections
f64_t = np.float64
i32_t = np.int32
u32_t = np.uint32


class Testrecarray2dict(unittest.TestCase):
    """Test the function that converts numpy record arrays into dicts for use within SatPy."""

    def test_fun(self):
        """Test record array."""
        inner_st = np.dtype([('test_str', '<S20'), ('test_int', 'i4')])
        outer_st = np.dtype([('test_sec', inner_st), ('test_flt', 'f4')])

        inner_da = np.array([('Testing', 10)], dtype=inner_st)
        outer_da = np.array([(inner_da, 1.45)], dtype=outer_st)

        expected = {'test_sec': {'test_str': np.array([b'Testing'], dtype='<S20'),
                                 'test_int': np.array([10], dtype=np.int32)},
                    'test_flt': np.array([1.45], dtype=np.float32)}
        self.assertEqual(expected, recarray2dict(outer_da))


class TestHRITGOMSProFileHandler(unittest.TestCase):
    """Test the HRIT Prologue FileHandler."""

    # Below are variable definitions used in testing the prologue reader.
    # These values are taken from a typical ELECTRO-L HRIT scene
    test_sat_status = {'TagType': 2,
                       'TagLength': 292,
                       'SatelliteID': 19002,
                       'SatelliteName': b'ELECTRO',
                       'NominalLongitude': 1.3264,
                       'SatelliteCondition': 1,
                       'TimeOffset': 0.}

    test_img_acq = {'TagType': np.repeat(3, 10).astype(u32_t),
                    'TagLength': np.repeat(24, 10).astype(u32_t),
                    'Status': np.repeat(2, 10).astype(u32_t),
                    'StartDelay': np.repeat(9119019, 10).astype(np.int32),
                    'Cel': np.repeat(0., 10)}

    test_calib = np.full((10, 1024), 50, dtype=np.int32)

    test_pro = {'SatelliteStatus': test_sat_status,
                'ImageAcquisition': test_img_acq,
                'ImageCalibration': test_calib}

    @mock.patch('satpy.readers.electrol_hrit.np.fromfile')
    @mock.patch('satpy.readers.electrol_hrit.HRITFileHandler.__init__')
    def test_init(self, new_fh_init, fromfile):
        """Set up the hrit file handler for testing."""
        new_fh_init.return_value.filename = 'filename'
        HRITGOMSPrologueFileHandler.filename = 'filename'
        HRITGOMSPrologueFileHandler.mda = {'total_header_length': 1}

        # Set up the test data to use within the prologue reader
        tss = np.array([(2, 292, 19002, 'ELECTRO', 1.3264, 1, 0.)],
                       dtype=satellite_status)
        tia = np.tile(np.array([(3, 24, 2, 9119019, 0.)],
                      dtype=image_acquisition), (1, 10))
        tc = np.full((10, 1024), 50)
        rtv = np.array([(tss, tia, tc)], dtype=prologue)
        # Pretend to return this when reading a fake prologue
        fromfile.return_value = rtv
        m = mock.mock_open()

        with mock.patch('satpy.readers.electrol_hrit.open',
                        m, create=True) as newopen:
            newopen.return_value.__enter__.return_value.seek.return_value = 1
            test_t = datetime.datetime(2018, 1, 1, 0, 0)
            self.reader = HRITGOMSPrologueFileHandler(
                             'filename', {'platform_shortname': 'GOMS2',
                                          'start_time': test_t,
                                          'service': 'test_service'},
                             {'filetype': 'info'})

        # assertDictEqual doesn't seem to work for dicts containing dicts,
        # so we must compare some items individually
        self.assertDictEqual(self.test_pro['SatelliteStatus'],
                             self.reader.prologue['SatelliteStatus'])
        prop = 'ImageAcquisition'
        for key in self.reader.prologue[prop]:
            np.testing.assert_array_equal(self.test_pro[prop][key],
                                          self.reader.prologue[prop][key])
        np.testing.assert_array_equal(self.test_pro['ImageCalibration'],
                                      self.reader.prologue['ImageCalibration'])


class TestHRITGOMSEpiFileHandler(unittest.TestCase):
    """Test the HRIT Epilogue FileHandler."""

    @mock.patch('satpy.readers.electrol_hrit.np.fromfile')
    @mock.patch('satpy.readers.electrol_hrit.HRITFileHandler.__init__')
    def test_init(self, new_fh_init, fromfile):
        """Set up the hrit file handler for testing."""
        new_fh_init.return_value.filename = 'filename'
        HRITGOMSEpilogueFileHandler.filename = 'filename'
        HRITGOMSEpilogueFileHandler.mda = {'total_header_length': 1}

        # Set up the test data to use within the epilogue reader
        rtv = np.ones((1,), dtype=epilogue)
        # Pretend to return this when reading a fake epilogue
        fromfile.return_value = rtv
        m = mock.mock_open()

        with mock.patch('satpy.readers.electrol_hrit.open',
                        m, create=True) as newopen:
            newopen.return_value.__enter__.return_value.seek.return_value = 1
            test_t = datetime.datetime(2018, 1, 1, 0, 0)
            self.reader = HRITGOMSEpilogueFileHandler(
                             'filename', {'platform_shortname': 'GOMS2',
                                          'start_time': test_t,
                                          'service': 'test_service'},
                             {'filetype': 'info'})

            epi = self.reader.epilogue

            # We don't check everything in the epilogue (too many nested dicts)
            # but rather check the epilogue is returned as a dict and that two
            # representative data fields are as we expect.
            self.assertIsInstance(epi, dict)
            np.testing.assert_array_equal(
                epi['RadiometricProcessing']['RPSummary']['IsOptic'], np.ones(10))
            np.testing.assert_array_equal(
                epi['GeometricProcessing']['TimeProcessing'], np.ones(10))


@mock.patch('satpy.readers.electrol_hrit.HRITGOMSFileHandler.__init__', return_value=None)
@mock.patch('satpy.readers.electrol_hrit.HRITFileHandler.get_dataset', return_value={})
class TestHRITGOMSFileHandler(unittest.TestCase):
    """A test of the ELECTRO-L main file handler functions."""

    @mock.patch('satpy.readers.electrol_hrit.HRITGOMSFileHandler.calibrate')
    def test_get_dataset(self, calibrate_mock, *mocks):
        """Test get dataset."""
        key = make_dataid(name='CH1', calibration='counts')
        fake_array = mock.MagicMock()
        fake_array.attrs = dict()
        calibrate_mock.return_value = fake_array
        fh = HRITGOMSFileHandler()
        fh.platform_name = 'Electro'
        fh.mda = {'projection_parameters': {'SSP_longitude': 0.0},
                  'orbital_parameters': {'satellite_nominal_longitude': 0.5}}
        info = {'units': 'm', 'standard_name': 'electro', 'wavelength': 5.0}
        output = fh.get_dataset(key, info)

        # Check that 'calibrate' is called
        calibrate_mock.assert_called()

        # Check that the correct attributes are returned
        attrs_exp = info.copy()
        attrs_exp.update({'orbital_parameters': {'satellite_nominal_longitude': 0.5,
                                                 'satellite_nominal_latitude': 0.0,
                                                 'projection_longitude': 0.0,
                                                 'projection_latitude': 0.0,
                                                 'projection_altitude': 35785831.00},
                          'platform_name': 'Electro',
                          'sensor': 'msu-gs'})
        assert dict(output.attrs, **attrs_exp) == output.attrs

    def test_calibrate(self, *mocks):
        """Test calibrate."""
        lut = np.linspace(1e6, 1.6e6, num=1024).astype(np.int32)
        lut = np.tile(lut, (10, 1))
        fh = HRITGOMSFileHandler()
        fh.prologue = {'ImageCalibration': lut}
        fh.chid = 1

        # Set up test input data

        counts = DataArray(da.linspace(1, 1023, 25, chunks=5,
                                       dtype=np.uint16).reshape(5, 5))

        # Test that calibration fails if given a silly mode
        self.assertRaises(NotImplementedError, fh.calibrate, counts,
                          'nonsense')

        # Test that 'counts' calibration returns identical values to input
        out = fh.calibrate(counts, 'counts')
        self.assertTrue(np.all(out.values == counts.values))

        # Test that 'radiance' calibrates successfully
        out = fh.calibrate(counts, 'radiance')
        self.assertTrue(np.allclose(out.values, lut[0, counts]/1000.))

        # Test that 'brightness_temperature' calibrates successfully
        out = fh.calibrate(counts, 'brightness_temperature')
        self.assertTrue(np.allclose(out.values, lut[0, counts]/1000.))

    def test_get_area_def(self, *mocks):
        """Test get_area_def."""
        example_area_ext = (-5566748.0802, -1854249.1809,
                            5570748.6178, 2000.2688)

        fh = HRITGOMSFileHandler()
        fh.mda = {'cfac': 10231753, 'lfac': 10231753,
                  'coff': 1392.0, 'loff': 0.0, 'number_of_lines': 464,
                  'number_of_columns': 2784,
                  'projection_parameters': {'SSP_longitude': 0.0}}
        area = fh.get_area_def(True)

        self.assertTrue(np.allclose(np.array(area.area_extent),
                        np.array(example_area_ext)))



#!/usr/bin/env python
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Enhancement functions specific to the ABI sensor."""

from satpy.enhancements import apply_enhancement


def cimss_true_color_contrast(img, **kwargs):
    """Scale data based on CIMSS True Color recipe for AWIPS."""
    def func(img_data):
        """Perform per-chunk enhancement.

        Code ported from Kaba Bah's AWIPS python plugin for creating the
        CIMSS Natural (True) Color image in AWIPS. AWIPS provides that python
        code the image data on a 0-255 scale. Satpy gives this function the
        data on a 0-1.0 scale (assuming linear stretching and sqrt
        enhancements have already been applied).

        """
        max_value = 1.0
        acont = (255.0 / 10.0) / 255.0
        amax = (255.0 + 4.0) / 255.0
        amid = 1.0 / 2.0
        afact = (amax * (acont + max_value) / (max_value * (amax - acont)))
        aband = (afact * (img_data - amid) + amid)
        aband[aband <= 10 / 255.0] = 0
        aband[aband >= 1.0] = 1.0
        return aband

    apply_enhancement(img.data, func, pass_dask=True)
